# -*- coding: utf-8 -*-
"""JSON File Wrappers."""

# Standard Library Imports
from __future__ import annotations
import abc
import json
import os
from typing import Any
from typing import IO
from typing import Optional
from typing import Union

# Local Imports
from .abstract_file_wrappers import AbstractDirectoryWrapper
from .abstract_file_wrappers import AbstractFileWrapper
from .abstract_file_wrappers import AbstractTextWrapper
from .abstract_file_wrappers import AbstractIOWrapper
from ..utils import converters
from .. import settings
from .. import utils

__all__ = ["JsonDirectoryWrapper", "JsonFileWrapper"]


class AbstractJsonWrapper(AbstractTextWrapper):
    """Represents an abstract wrapper class for `.json` files."""

    @property
    @abc.abstractmethod
    def indent(self) -> Optional[Union[int, str]]:
        """Indent."""
        raise NotImplementedError

    def _init_json_io_wrapper(self, __file: IO[Any], /) -> JsonIOWrapper:
        """Initialize I/O wrapper for `.json` file.

        Args:
            __file: File-like object.

        Returns:
            I/O wrapper instance.

        """
        result = JsonIOWrapper(__file)
        setattr(result, "_context", self)
        return result


class JsonDirectoryWrapper(AbstractJsonWrapper, AbstractDirectoryWrapper):
    """Implements a wrapper for `.json` files in a directory.

    Args:
        directory: Directory from which to load `.json` file(s).
        encoding (optional): File encoding. Default `utf-8`.
        indent (optional): Indent. Default ``None``.
        read_only (optional): Whether file is read only. Default ``False``.

    """

    def __init__(
        self,
        directory: "os.PathLike[Any]",
        *,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        indent: Optional[Union[int, str]] = None,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            directory,
            encoding=encoding,
            extension=settings.JSON_EXTENSION,
            read_only=read_only,
        )
        self._indent = indent

    @property
    def indent(self) -> Optional[Union[int, str]]:
        """Indent."""
        return self._indent

    def open(self, filename: str, /, mode: str = "r") -> JsonIOWrapper:
        """Open a `.json` file and return a file object.

        Args:
            filename: Filename.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        Raises:
            TypeError: when `encoding` is not type ``str``.

        """

        file = super().open(filename, mode=converters.to_text_file_mode(mode))
        result = self._init_json_io_wrapper(file)
        return result


class JsonFileWrapper(AbstractJsonWrapper, AbstractFileWrapper):
    """Implements a wrapper for `.json` files.

    Args:
        filepath: Path to `.json` file.
        encoding (optional): File encoding. Default `utf-8`.
        indent (optional): Indent. Default ``None``.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        ValueError: when `filepath` is not a `.json` file.

    """

    def __init__(
        self,
        filepath: "os.PathLike[Any]",
        *,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        indent: Optional[Union[int, str]] = None,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            filepath,
            encoding=encoding,
            read_only=read_only,
        )
        utils.raise_for_extension(filepath, settings.JSON_EXTENSION)

        self._indent = indent

    @property
    def indent(self) -> Optional[Union[int, str]]:
        """Indent."""
        return self._indent

    def open(self, mode: str = "r") -> JsonIOWrapper:
        """Open the `.json` file and return a file object.

        Args:
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        """
        file = super().open(converters.to_text_file_mode(mode))
        result = self._init_json_io_wrapper(file)
        return result


class JsonIOWrapper(AbstractIOWrapper):
    """Implements a I/O wrapper for `.json` files."""

    def __init__(self, __file: IO[Any]) -> None:
        self._file = __file
        self._context: Optional[AbstractJsonWrapper] = None

    @property
    def file(self) -> IO[Any]:
        """File."""
        return self._file

    @property
    def closed(self) -> bool:
        """Whether file is closed."""
        return self._file.closed

    @property
    def indent(self) -> Optional[Union[int, str]]:
        """Indent."""
        default = getattr(self._context, "indent", None)
        result = getattr(self, "_indent", default)
        return result

    @indent.setter
    def indent(self, value: Any) -> None:
        if not isinstance(value, (int, str)):
            expected = "expected type 'int' or 'str'"
            actual = f"got {type(value)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        setattr(self, "_indent", value)

    @property
    def read_only(self) -> bool:
        """Whether read only."""
        return getattr(self._context, "read_only")  # type: bool

    def __enter__(self) -> JsonIOWrapper:
        return self

    def __exit__(self, *_) -> None:
        self.close()
        return

    def close(self) -> None:
        """Close `.csv` file."""
        self._file.close()

    def dump(self, obj: Any) -> None:
        """Serialize `obj` to file."""
        json.dump(obj, self._file)

    def load(self) -> Any:
        """Deserialize contents of file."""
        result = json.load(self._file)
        return result



# -*- coding: utf-8 -*-

# Standard Library Imports
import importlib

# Local Imports
from .abstract_progress_bar import *

try:
    importlib.import_module("tqdm")
except (ImportError, ModuleNotFoundError):
    pass
else:
    from .tqdm_progress_bar import *



# -*- coding: utf-8 -*-
"""Timestamp Converter.

Module provides function for converting values to timestamps.

"""

# Standard Library Imports
import datetime
import decimal
import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional

# Third-Party Imports
from dateutil.parser import parse
from dateutil.parser import ParserError

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_timestamp"]


def to_timestamp(__value: Any, /, default: float = 0.0) -> float:
    """Convert date to timestamp.

    Args:
        __value: Value to convert to timestamp.
        default (optional): Default value. Default ``0.0``.

    Returns:
        Timestamp.
    """
    converter = TimestampConverter(default=default)
    result = converter(__value)
    return result


class TimestampConverter(BaseConverter):
    """Class implements a timestamp converter.

    Args:
        default (optional): Default value. Default ``0.0``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: float = 0.0,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if not isinstance(default, float):
            message = f"expected type 'float', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()


def timestamp_from_date(
    __value: datetime.date, _: Optional[float] = None, /
) -> float:
    """Convert date value to timestamp.

    Args:
        __value: Value to convert to timestamp.

    Returns:
        Timestamp.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    result = time.mktime(__value.timetuple())
    return result


def timestamp_from_datetime(
    __value: datetime.datetime, _: Optional[float] = None, /
) -> float:
    """Convert datetime value to timestamp.

    Args:
        __value: Value to convert to timestamp.

    Returns:
        Timestamp.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    result = __value.timestamp()
    return result


def timestamp_from_decimal(
    __value: decimal.Decimal, _: Optional[float] = None, /
) -> float:
    """Convert decimal value to timestamp.

    Args:
        __value: Value to convert to timestamp.

    Returns:
        Timestamp.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def timestamp_from_float(
    __value: float, _: Optional[float] = None, /
) -> float:
    """Convert float value to timestamp.

    Args:
        __value: Value to convert to timestamp.

    Returns:
        Timestamp.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def timestamp_from_int(__value: int, _: Optional[float] = None, /) -> float:
    """Convert integer value to timestamp.

    Args:
        __value: Value to convert to timestamp.

    Returns:
        Timestamp.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def timestamp_from_str(
    __value: str, default: Optional[float] = None, /
) -> Optional[float]:
    """Convert string value to timestamp.

    Args:
        __value: Value to convert to timestamp.
        default (optional): Default value. Default ``None``.

    Returns:
        Timestamp.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to timestamp.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    try:
        value = __value.replace("  ", " ").strip()
        result = parse(value).timestamp() if value else default

    except (ParserError, ValueError) as error:
        message = f"'{__value}' cannot be converted to timestamp"
        raise ValueError(message) from error

    return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., Optional[float]]] = {
    datetime.date: timestamp_from_date,
    datetime.datetime: timestamp_from_datetime,
    decimal.Decimal: timestamp_from_decimal,
    float: timestamp_from_float,
    int: timestamp_from_int,
    str: timestamp_from_str,
}



# -*- coding: utf-8 -*-
"""Progressive Unit of Work."""

# Standard Library Imports
from __future__ import annotations
from typing import Any

# Local Imports
from .abstract_unit_of_work import AbstractUnitOfWork
from ..progress import AbstractProgressBar

__all__ = ["ProgressiveUnitOfWork"]


class ProgressiveUnitOfWork(AbstractUnitOfWork):
    """Class implements a progressive unit of work.

    Args:
        *args (optional): Positional arguments.
        progress_bar: Progress bar.
        **kwargs (optional): Keyword arguments.

    Attributes:
        progress_bar: Progress bar.

    Raises:
        TypeError: when `progress_bar` parameter is not type 'tqdm'.

    """

    def __init__(
        self,
        *args: Any,
        progress_bar: AbstractProgressBar,
        **kwargs: Any,
    ) -> None:
        if not isinstance(progress_bar, AbstractProgressBar):  # type: ignore
            expected = "expected type 'ProgressBar'"
            actual = f"got {type(progress_bar)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        super().__init__(*args, **kwargs)
        self._progress_bar = progress_bar

    def __exit__(self, *args: Any) -> None:
        super().__exit__(*args)

        if self._progress_bar.leave is False:
            self._progress_bar.close()

    @property
    def progress(self) -> AbstractProgressBar:
        """Progress bar."""
        return self._progress_bar



# -*- coding: utf-8 -*-
"""Abstract Repository.

Implementation based on 'Architecture Patterns in Python' repository pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# Standard Library Imports
import abc
from typing import Any
from typing import List
from typing import Set
from typing import Union

# Local Imports
from ..metaclasses import RepositoryMeta
from ..metaclasses.tracker import SEEN_ATTR

__all__ = ["AbstractRepository"]


class AbstractRepository(abc.ABC, metaclass=RepositoryMeta):
    """Represents an abstract repository."""

    @property
    def seen(self) -> Set[Any]:
        """Objects seen."""
        return getattr(self, SEEN_ATTR, set())

    @abc.abstractmethod
    def add(self, obj: Any) -> None:
        """Add object to repository.

        Args:
            obj: Object to add to repository.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, ref: Union[int, str]) -> Any:
        """Get object from repository.

        Args:
            ref: Reference to object.

        Returns:
            Object.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def list(self) -> List[Any]:
        """List objects in repository.

        Returns:
            Objects in repository.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, obj: Any) -> None:
        """Remove object from repository.

        Args:
            obj: Object to remove from repository.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def commit(self) -> None:
        """Commit changes to repository."""
        raise NotImplementedError

    @abc.abstractmethod
    def rollback(self) -> None:
        """Rollback changes to repository."""
        raise NotImplementedError



# -*- coding: utf-8 -*-

# Standard Library Imports
from __future__ import annotations
from typing import List
from typing import Optional
from typing import Set
from typing import Union

# Third-Party Imports
from ..models import AbstractModel
from ..repositories import AbstractRepository
from ..repositories import EventfulRepository

__all__ = ["FakeRepository", "FakeEventfulRepository"]


# Constants
DEFAULT_KEY = "reference"

# Private Attributes
CLOSED_ATTR = "_closed"
COMMITTED_ATTR = "_committed"
ROLLED_BACK_ATTR = "_rolled_back"


class FakeRepository(AbstractRepository):
    """Implements a fake repository."""

    def __init__(
        self,
        objects: Optional[List[AbstractModel]] = None,
        key: str = DEFAULT_KEY,
    ) -> None:
        super().__init__()
        self._objects: Set[AbstractModel] = set(objects or [])
        self._key = key

    @property
    def closed(self) -> bool:
        """Whether `close()` method was called."""
        result = getattr(self, CLOSED_ATTR, False)
        return result

    @property
    def committed(self) -> bool:
        """Whether `commit()` method was called."""
        result = getattr(self, COMMITTED_ATTR, False)
        return result

    @property
    def rolled_back(self) -> bool:
        """Whether `rollback()` method was called."""
        result = getattr(self, ROLLED_BACK_ATTR, False)
        return result

    def __contains__(self, obj: AbstractModel) -> bool:
        return obj in self._objects

    def add(self, obj: AbstractModel) -> None:
        """Add object."""
        self._objects.add(obj)

    def get(self, ref: Union[int, str]) -> Optional[AbstractModel]:
        """Get object.

        Args:
            ref: Reference to object.

        """
        results = (
            obj
            for obj in self._objects
            if getattr(obj, self._key, None) == ref
        )
        return next(results, None)

    def list(self) -> List[AbstractModel]:
        """List objects."""
        return list(self._objects)

    def remove(self, obj: AbstractModel) -> None:
        """Remove object."""
        self._objects.discard(obj)

    def commit(self) -> None:
        """Commit changes."""
        setattr(self, COMMITTED_ATTR, True)

    def rollback(self) -> None:
        """Rollback changes."""
        setattr(self, ROLLED_BACK_ATTR, True)

    def close(self) -> None:
        """Close repository."""
        setattr(self, CLOSED_ATTR, True)


class FakeEventfulRepository(FakeRepository, EventfulRepository):
    """Implements a fake eventful repository."""



# -*- coding: utf-8 -*-
"""Abstract Converter.

Module defines an abstract class for converting values between data types.

"""

# Standard Library Imports
import abc
from typing import Any
from typing import Callable

__all__ = ["AbstractConverter"]


class AbstractConverter(abc.ABC):
    """Class represents an abstract converter."""

    @abc.abstractmethod
    def __call__(self, __value: Any, /) -> Any:
        raise NotImplementedError

    @abc.abstractmethod
    def set_conversion(
        self, __type: type, func: Callable[..., Any], /
    ) -> None:
        """Set conversion for type.

        Args:
            __type: Type.
            func: Function for converting between types.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def reset(self) -> None:
        """Reset converter."""
        raise NotImplementedError



# -*- coding: utf-8 -*-

# Standard Library Imports
import importlib

# Local Imports
from .abstract_dispatcher import *
from .base_dispatcher import *

try:
    importlib.import_module("tqdm")
except (ImportError, ModuleNotFoundError):
    pass
else:
    from .progressive_dispatcher import *



# -*- coding: utf-8 -*-
"""Queues.

This module defines classes for queues.

"""

# pylint: disable=too-few-public-methods

# Standard Library Imports
from __future__ import annotations
import abc
from collections import deque
from typing import Any
from typing import Deque
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Type
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Self

# Local Imports
from .messages import AbstractMessage
from .helpers import raise_for_instance

__all__ = [
    "AbstractQueue",
    "MessageQueue",
]


class AbstractQueue(abc.ABC):
    """Represents an abstract queue.

    Args:
        __iterable (optional): Iterable of objects. Default ``None``.

    """

    def __init__(
        self: Self,
        __iterable: Optional[Iterable[object]] = None,
    ) -> None:
        raise_for_iterable(__iterable)

        self._items: Deque[Any] = deque(__iterable or [])
        self.sort()

    def __iter__(self) -> Iterator[object]:
        return self

    def __len__(self) -> int:
        return len(self._items)

    def __next__(self) -> Any:
        if len(self._items) == 0:
            raise StopIteration

        return self.popleft()

    def __repr__(self) -> str:
        return repr(list(self._items))

    def append(self, __item: object, /) -> None:
        """Append item to queue.

        Args:
            __item: Item to append to queue.

        """
        self._items.append(__item)
        self.sort()

    def extend(self, __iterable: Iterable[object], /) -> None:
        """Extend queue.

        Args:
            __iterable: Items with which to extend queue.

        """
        self._items.extend(__iterable)
        self.sort()

    def popleft(self) -> object:
        """Pop item from left side of queue."""
        return self._items.popleft()

    def sort(self) -> None:
        """Sort queue."""
        self._items = deque(sorted(self._items))

    def clear(self) -> None:
        """Clear queue."""
        self._items.clear()


class MessageQueue(AbstractQueue):
    """Implements a message queue.

    Args:
        __iterable (optional): Iterable of messages. Default ``None``.

    """

    def __new__(
        cls: Type[Self],
        __iterable: Optional[Iterable[object]] = None,
    ) -> MessageQueue:
        if __iterable is not None:
            for item in __iterable:
                raise_for_instance(item, AbstractMessage)

        instance = super().__new__(cls)
        return instance

    def __iter__(self) -> Iterator[AbstractMessage]:
        return self

    def __next__(self) -> AbstractMessage:
        return super().__next__()

    def append(self, __item: object, /) -> None:
        """Append message to queue.

        Args:
            __item: Message to append to queue.

        """
        raise_for_instance(__item, AbstractMessage)
        super().append(__item)

    def extend(self, __iterable: Iterable[object], /) -> None:
        """Extend queue with messages.

        Args:
            __iterable: Messages with which to extend queue.

        """
        messages: List[AbstractMessage] = []
        for item in __iterable:
            raise_for_instance(item, AbstractMessage)
            messages.append(item)  # type: ignore

        super().extend(messages)

    def popleft(self) -> AbstractMessage:
        """Pop message from left side of queue."""
        return super().popleft()  # type: ignore


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
def raise_for_iterable(__obj: object) -> None:
    """Raise error when object is not iterable.

    Args:
        __obj: Object to check.

    Raises:
        TypeError: when object is not iterable.

    """
    if __obj and not isinstance(__obj, Iterable):
        raise TypeError("argument is not iterable")



# -*- coding: utf-8 -*-
"""TXT File Wrappers."""

# Standard Library Imports
from __future__ import annotations
import logging
import io
import os
from typing import Any
from typing import Iterable
from typing import List
from typing import Optional

# Local Imports
from .abstract_file_wrappers import AbstractDirectoryWrapper
from .abstract_file_wrappers import AbstractFileWrapper
from .abstract_file_wrappers import AbstractTextWrapper
from ..utils import converters
from .. import settings
from .. import utils

__all__ = [
    "TxtDirectoryWrapper",
    "TxtFileWrapper",
]


# Initiate logger.
log = logging.getLogger("dodecahedron")


class AbstractTxtWrapper(AbstractTextWrapper):
    """Represents an abstract wrapper class for `.txt` files."""

    def _init_txt_io_wrapper(
        self,
        __buffer: memoryview,
        /,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        newline: str = settings.DEFAULT_TXT_NEWLINE,
    ) -> TxtIOWrapper:
        """Initialize I/O wrapper for `.txt` file.

        Args:
            __buffer: Buffer.

        Returns:
            I/O wrapper instance.

        """
        result = TxtIOWrapper(
            __buffer,
            encoding=encoding,
            newline=newline,
        )
        setattr(result, "_context", self)
        return result


class TxtDirectoryWrapper(AbstractTxtWrapper, AbstractDirectoryWrapper):
    """Implements a wrapper for `.txt` files in a directory.

    Args:
        directory: Directory from which to load `.txt` file(s).
        encoding (optional): File encoding. Default `utf-8`.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        TypeError: when `directory` is not type ``Path``.
        TypeError: when `encoding` is not type ``str``.
        NotADirectoryError: when `directory` is not a valid directory.

    """

    def __init__(
        self,
        directory: "os.PathLike[Any]",
        *,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        newline: Optional[str] = None,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            directory,
            encoding=encoding,
            extension=settings.TXT_EXTENSION,
            newline=newline,
            read_only=read_only,
        )

    def open(self, filename: str, /, mode: str = "r") -> TxtIOWrapper:
        """Open a `.txt` file and return a file object.

        Args:
            filename: Filename.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        Raises:
            TypeError: when `encoding` is not type ``str``.

        """

        file = super().open(filename, mode=converters.to_bytes_file_mode(mode))
        result = self._init_txt_io_wrapper(file)
        return result


class TxtFileWrapper(AbstractTxtWrapper, AbstractFileWrapper):
    """Implements a wrapper for `.txt` files.

    Args:
        filepath: Path to `.txt` file.
        encoding (optional): File encoding. Default `utf-8`.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        ValueError: when `filepath` is not a `.txt` file.

    """

    def __init__(
        self,
        filepath: "os.PathLike[Any]",
        *,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        newline: Optional[str] = None,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            filepath,
            encoding=encoding,
            newline=newline,
            read_only=read_only,
        )
        utils.raise_for_extension(filepath, settings.TXT_EXTENSION)

    def open(self, mode: str = "r") -> TxtIOWrapper:
        """Open the `.txt` file and return a file object.

        Args:
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        """
        file = super().open(converters.to_bytes_file_mode(mode))
        result = self._init_txt_io_wrapper(file)
        return result


class TxtIOWrapper(io.TextIOWrapper):
    """Implements a I/O wrapper for `.txt` files."""

    def __init__(
        self,
        buffer: memoryview,
        encoding: Optional[str] = None,
        errors: Optional[str] = None,
        newline: Optional[str] = None,
        line_buffering: bool = False,
        write_through: bool = False,
    ) -> None:
        super().__init__(
            buffer,
            encoding=encoding,
            errors=errors,
            newline=newline,
            line_buffering=line_buffering,
            write_through=write_through,
        )
        self._context: Optional[AbstractTxtWrapper] = None

    def read(self, size: Optional[int] = -1) -> str:
        """Read content of `.txt` file.

        Returns:
            Content.

        """
        result = super().read(size)
        return result

    def readline(self, size: int = -1) -> str:
        """Read line from `.txt` file.

        Returns:
            Line.

        """
        result = super().readline(size)
        return result

    def readlines(self, hint: int = -1) -> List[str]:
        """Read lines from `.txt` file.

        Returns:
            Lines.

        """
        result = super().readlines(hint)
        return result

    def write(self, content: str, /) -> int:
        """Write content to `.txt` file.

        Args:
            content: Content.

        """
        result = super().write(content)
        return result

    def writelines(self, lines: Iterable[str], /) -> None:
        """Write lines to `.txt` file.

        Args:
            lines: Lines.

        """
        result = super().writelines(lines)
        return result



# -*- coding: utf-8 -*-
"""Eventful Unit of Work.

Based on 'Architecture Patterns in Python' unit-of-work pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# Standard Library Imports
from __future__ import annotations
from typing import Any
from typing import Generator

# Local Imports
from .abstract_unit_of_work import AbstractUnitOfWork
from ..messages import AbstractMessage
from ..queues import MessageQueue

__all__ = ["EventfulUnitOfWork"]


class EventfulUnitOfWork(AbstractUnitOfWork):
    """Class implements an eventful unit of work."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._events = MessageQueue()

    def __enter__(self) -> EventfulUnitOfWork:
        self._events.clear()
        super().__enter__()
        return self

    @property
    def events(self) -> MessageQueue:
        """Event queue."""
        return self._events

    def collect_events(self) -> Generator[AbstractMessage, None, None]:
        """Collect events.

        Yields:
            Event.

        """
        while self.events:
            yield self.events.popleft()



# -*- coding: utf-8 -*-
"""Abstract Mapper."""

# Standard Library Imports
from __future__ import annotations
import abc
import collections
from typing import Any
from typing import Callable
from typing import Dict
from typing import Hashable
from typing import Literal
from typing import Optional
from typing import Union

__all__ = ["AbstractMapper"]


# Constants
INWARD = "inward"
OUTWARD = "outward"


class AbstractMapper(abc.ABC):
    """Represents an abstract mapper."""

    def __init__(self, __schema: Dict[Hashable, Any], /) -> None:
        self._schema = _MapperSchema(__schema)

    @property
    def schema(self) -> _MapperSchema:
        """Mapped schema."""
        return self._schema

    def _get_attribute_mapper(
        self, ref: Union[int, str]
    ) -> Optional[Dict[str, Any]]:
        """Get attribute mapper.

        Args:
            ref: Reference for attribute.

        Returns:
            Attribute mapper.

        """
        result = self._schema.get_attribute_mapper(ref)
        return result

    def _get_converter(
        self,
        ref: Union[int, str],
        direction: Literal["inward", "outward"] = OUTWARD,
        *,
        default: Callable[..., Any] = lambda x: x,  # type: ignore
    ) -> Optional[Callable[..., Any]]:
        """Get converter.

        Args:
            ref: Reference for attribute.
            direction (optional): Direction of converter. Default ``outward``.
            default (optional): Default converter. Default `lambda x: x`.

        Returns:
            Converter.

        """
        result = self._schema.get_converter(ref, direction, default=default)
        return result

    def _get_attribute_names(self) -> Dict[Hashable, str]:
        """Get attribute names.

        Returns:
            Attribute names.

        """
        results = self._schema.get_attribute_names()
        return results


class _MapperSchema(collections.UserDict):  # type: ignore
    """Implements a mapper schema."""

    def __init__(self, __mappers: Dict[Hashable, Any]) -> None:
        if not isinstance(__mappers, dict):  # type: ignore
            message = f"expected type 'dict', got {type(__mappers)} instead"
            raise TypeError(message)

        self._data = standardize_attribute_mappers(__mappers)

    @property
    def data(self) -> Dict[Hashable, Any]:  # type: ignore
        """Schema data."""
        return self._data

    def get_converter(
        self,
        ref: Union[int, str],
        direction: Literal["inward", "outward"] = OUTWARD,
        *,
        default: Callable[..., Any] = lambda x: x,  # type: ignore
    ) -> Optional[Callable[..., Any]]:
        """Get converter.

        Args:
            ref: Reference for attribute.
            direction (optional): Direction of converter. Default ``outward``.
            default (optional): Default converter. Default `lambda x: x`.

        Returns:
            Converter.

        """
        try:
            mapper = self.get_attribute_mapper(ref)
            result = mapper["converters"][direction] if mapper else None

        except KeyError:
            result = (
                mapper.get("converter") or default  # type: ignore
                if direction == OUTWARD
                else default
            )

        return result

    def get_attribute_mapper(
        self, ref: Union[int, str]
    ) -> Optional[Dict[str, Any]]:
        """Get attribute mapper.

        Args:
            ref: Reference for attribute.

        Returns:
            Attribute mapper.

        """
        try:
            result = self.data[ref]
        except KeyError:
            return None

        return result

    def get_attribute_names(self) -> Dict[Hashable, str]:
        """Get attribute names.

        Returns:
            Attribute names.

        """
        results: Dict[Hashable, str] = {
            get_attribute_name(attr): key
            for key, attr in self.data.items()
            if get_attribute_name(attr) is not None
        }  # type: ignore
        return results


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
def standardize_attribute_mappers(
    __mappers: Dict[Hashable, Any], /
) -> Dict[Hashable, Any]:
    """Standardize attribute mappers.

    Args:
        __mappers: Attribute mappers.

    Returns:
        Attribute mappers.

    """
    results = {
        key: standardize_attribute_mapper(value)
        for key, value in __mappers.items()
    }
    return results


def standardize_attribute_mapper(__mapper: Any, /) -> Dict[Hashable, Any]:
    """Standardize attribute mapper.

    Args:
        __mapper: Attribute mapper.

    Returns:
        Attribute mapper.

    """
    if isinstance(__mapper, dict):
        return __mapper  # type: ignore

    if isinstance(__mapper, str):
        return {"map_to": __mapper}

    message = "invalid attribute mapper"
    raise ValueError(message)


# ----------------------------------------------------------------------------
# Selectors
# ----------------------------------------------------------------------------
def get_attribute_name(__mapper: Any, /) -> Optional[str]:
    """Get mapped attribute.

    Args:
        __mapper: Attribute mapper.

    Returns:
        Mapped attribute.

    """
    if isinstance(__mapper, dict):
        return __mapper.get("map_to")  # type: ignore

    if isinstance(__mapper, str):
        return __mapper

    return None



# -*- coding: utf-8 -*-
"""Datetime Converter.

Module provides function for converting values to datatimes.

"""

# Standard Library Imports
import datetime
import decimal
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional
from typing import Union

# Third-Party Imports
import cachetools
from cachetools.keys import hashkey
from dateutil.parser import parse
from dateutil.parser import ParserError
from dateutil.tz import tzlocal
import pytz

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_datetime"]


def to_datetime(
    __value: Any,
    /,
    default: Optional[datetime.datetime] = None,
    timezone: Optional[Union[str, datetime.tzinfo]] = tzlocal(),
) -> Optional[datetime.datetime]:
    """Convert value to ``datetime``.

    Args:
        __value: Value to convert to ``datetime``.
        default (optional): Default value. Default ``None``.
        timezone (optional): Timezone. Default `local`.

    Returns:
        Datetime.

    """
    converter = DatetimeConverter(default=default, timezone=timezone)
    result = converter(__value)
    return result


class DatetimeConverter(BaseConverter):
    """Class implements a datetime converter.

    Args:
        default (optional): Default value. Default ``None``.
        timezone (optional): Timezone. Default ``None``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: Optional[datetime.datetime] = None,
        timezone: Optional[Union[str, datetime.tzinfo]] = None,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if default and not isinstance(default, datetime.datetime):  # type: ignore
            message = f"expected type 'datetime', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()
        self._timezone = timezone

    def __call__(self, __value: Any, /) -> Optional[datetime.datetime]:
        dt = super().__call__(__value)
        result = self._add_timezone(dt)
        return result

    def _add_timezone(
        self, __datetime: Optional[datetime.datetime]
    ) -> Optional[datetime.datetime]:
        """Add timezone.

        Args:
            __datetime: Datetime.

        Returns:
            Datetime.

        """
        if not __datetime or self._timezone is None:
            return __datetime

        if isinstance(self._timezone, str) and is_naive(__datetime):
            result = pytz.timezone(self._timezone).localize(__datetime)
            return result

        result = __datetime.astimezone(self._timezone)  # type: ignore
        return result


def datetime_from_date(
    __value: datetime.date, _: Optional[datetime.datetime], /
) -> datetime.datetime:
    """Convert date value to ``datetime``.

    Args:
        __value: Value to convert to ``datetime``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    time = datetime.datetime.min.time()
    result = datetime.datetime.combine(__value, time)
    return result


def datetime_from_datetime(
    __value: datetime.datetime, _: Optional[datetime.datetime], /
) -> datetime.datetime:
    """Convert datetime value to ``datetime``.

    Args:
        __value: Value to convert to ``datetime``.

    Returns:
        Datetime.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    result = datetime.datetime(
        __value.year,
        __value.month,
        __value.day,
        __value.hour,
        __value.minute,
        __value.second,
        __value.microsecond,
        __value.tzinfo,
    )
    return result


def datetime_from_decimal(
    __value: decimal.Decimal, _: Optional[datetime.datetime], /
) -> datetime.datetime:
    """Convert decimal value to ``datetime``.

    Args:
        __value: Value to convert to ``datetime``.

    Returns:
        Datetime.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = datetime_from_float(float(__value), _)
    return result


def datetime_from_float(
    __value: float, _: Optional[datetime.datetime], /
) -> datetime.datetime:
    """Convert float value to ``datetime``.

    Args:
        __value: Value to convert to ``datetime``.

    Returns:
        Datetime.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    try:
        result = datetime_from_serial_date(__value, _)
    except ValueError:
        result = datetime_from_timestamp(__value, _)

    return result


def datetime_from_serial_date(
    __value: float, _: Optional[datetime.datetime], /
) -> datetime.datetime:
    """Convert serial date value to ``datetime``.

    Implementation based on reply to a question on Stack Overflow.

    Args:
        __value: Value to convert to ``datetime``.

    Returns:
        Datetime.

    Raises:
        TypeError: when value is not type 'float'.

    .. _Stack Overflow:
        https://stackoverflow.com/a/6706556/16732779

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    seconds = (__value - 25569) * 86400.0
    result = datetime_from_timestamp(seconds, _)
    return result


def datetime_from_int(
    __value: int, _: Optional[datetime.datetime], /
) -> datetime.datetime:
    """Convert integer value to ``datetime``.

    Args:
        __value: Value to convert to ``datetime``.

    Returns:
        Datetime.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = datetime_from_timestamp(float(__value), _)
    return result


def datetime_from_timestamp(
    __value: Union[float, int], _: Optional[datetime.datetime], /
) -> datetime.datetime:
    """Convert timestamp value to ``datetime``.

    Args:
        __value: Value to convert to ``datetime``.

    Returns:
        Datetime.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = datetime.datetime.fromtimestamp(__value)
    return result


@cachetools.cached(cachetools.LRUCache(maxsize=1000), hashkey)
def datetime_from_str(
    __value: str, default: Optional[datetime.datetime] = None, /
) -> Optional[datetime.datetime]:
    """Convert string value to ``datetime``.

    Args:
        __value: String representation of datetime.
        default (optional): Default value. Default ``None``.

    Returns:
        Datetime.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to ``datetime``.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    try:
        value = __value.replace("  ", " ").strip()
        result = parse(value) if value else default

    except (ParserError, ValueError) as error:
        message = f"'{__value}' cannot be converted to datetime"
        raise ValueError(message) from error

    return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., Optional[datetime.datetime]]] = {
    datetime.date: datetime_from_date,
    datetime.datetime: datetime_from_datetime,
    decimal.Decimal: datetime_from_decimal,
    float: datetime_from_float,
    int: datetime_from_int,
    str: datetime_from_str,
}


# ----------------------------------------------------------------------------
# Validators
# ----------------------------------------------------------------------------
def is_naive(__dt: datetime.datetime) -> bool:
    """Check whether datetime is naive.

    Args:
        __dt: Datetime.

    Returns:
        Wether datetime is naive.

    """
    result = __dt.tzinfo is None or __dt.tzinfo.utcoffset(__dt) is None
    return result



# -*- coding: utf-8 -*-
"""Xlsx File Wrappers."""

# Standard Library Imports
from __future__ import annotations
import os
from typing import Any
from typing import Dict
from typing import IO
from typing import Iterable
from typing import Iterator
from typing import List
from typing import Optional
from typing import Sequence
from typing import Tuple
from typing import TypeVar

# Third-Party Imports
from openpyxl import load_workbook
from openpyxl import Workbook
from openpyxl.worksheet.worksheet import Worksheet

# Local Imports
from .abstract_file_wrappers import AbstractDirectoryWrapper
from .abstract_file_wrappers import AbstractFileSystemWrapper
from .abstract_file_wrappers import AbstractFileWrapper
from .abstract_file_wrappers import AbstractIOWrapper
from ..utils import converters
from .. import settings
from .. import utils

__all__ = [
    "XlsxDirectoryWrapper",
    "XlsxFileWrapper",
    "XlsxIOWrapper",
]


# Custom type
T = TypeVar("T")


class AbstractXlsxWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for `.xlsx` files."""

    def _init_xlsx_io_wrapper(
        self, __file: IO[Any], /, mode: str
    ) -> XlsxIOWrapper:
        """Initialize I/O wrapper for `.xlsx` file.

        Args:
            __file: File-like object.
            mode: File mode.

        Returns:
            I/O wrapper instance.

        """
        result = XlsxIOWrapper(__file, mode=mode)
        setattr(result, "_context", self)
        return result


class XlsxDirectoryWrapper(AbstractXlsxWrapper, AbstractDirectoryWrapper):
    """Implements a wrapper for `.xlsx` files in a directory.

    Args:
        directory: Directory from which to load `.xlsx` file(s).
        fieldnames (optional): Fieldnames. Default ``None``.
        read_only (optional): Whether file is read only. Default ``False``.

    """

    def __init__(
        self,
        directory: "os.PathLike[Any]",
        *,
        fieldnames: Optional[Sequence[str]] = None,
        read_only: bool = False,
    ) -> None:
        super().__init__(
            directory,
            extension=settings.XLSX_EXTENSION,
            read_only=read_only,
        )
        self._fieldnames = fieldnames

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames or []

    def open(self, filename: str, /, mode: str = "rb") -> XlsxIOWrapper:
        """Open a `.xlsx` file and return a file object.

        Args:
            filename: Filename.
            mode (optional): File mode. Default ``rb``.

        Returns:
            File object.

        Raises:
            TypeError: when `encoding` is not type ``str``.

        """

        file = super().open(filename, mode=converters.to_bytes_file_mode(mode))
        result = self._init_xlsx_io_wrapper(file, mode=mode)
        return result


class XlsxFileWrapper(AbstractXlsxWrapper, AbstractFileWrapper):
    """Implements a wrapper for `.xlsx` files.

    Args:
        filepath: Path to `.xlsx` file.
        fieldnames (optional): Fieldnames. Default ``None``.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        ValueError: when `filepath` is not a `.xlsx` file.

    """

    def __init__(
        self,
        filepath: "os.PathLike[Any]",
        *,
        fieldnames: Optional[Sequence[str]] = None,
        read_only: bool = False,
    ) -> None:
        super().__init__(filepath, read_only=read_only)
        utils.raise_for_extension(filepath, settings.XLSX_EXTENSION)

        self._fieldnames = fieldnames or []

    @property
    def fieldnames(self) -> Sequence[Any]:
        """Fieldnames."""
        return self._fieldnames

    def open(self, mode: str = "rb") -> XlsxIOWrapper:
        """Open the `.xlsx` file and return a file object.

        Args:
            mode (optional): File mode. Default ``rb``.

        Returns:
            File object.

        """
        file = super().open(converters.to_bytes_file_mode(mode))
        result = self._init_xlsx_io_wrapper(file, mode=mode)
        return result


class OpenPyXLMixin:
    """Implements an OpenPyXL mixin."""

    _file: IO[Any]
    _mode: str

    def __exit__(self, *_: Any) -> None:
        read_only: bool = getattr(self, "read_only")
        if "w" in self._mode and not read_only:
            self.save()

    def get_worksheet(self) -> Worksheet:
        """Get worksheet.

        Returns:
            Worksheet.

        """
        workbook = self.get_workbook()
        sheetname = getattr(self, "sheetname", None)
        result = workbook[sheetname] if sheetname else workbook.active
        return result  # type: ignore

    def get_workbook(self) -> Workbook:
        """Get workbook.

        Returns:
            Workbook.

        """
        result = getattr(self, "_workbook", None)
        if not result and "r" in self._mode:
            result = self.load_workbook()

        if not result and "w" in self._mode:
            result = self.make_workbook()

        return result  # type: ignore

    def load_workbook(self) -> Workbook:
        """Load workbook.

        Returns:
            Workbook.

        """
        read_only: bool = getattr(self, "read_only")
        result = load_workbook(self._file, read_only=read_only)
        setattr(self, "_workbook", result)
        return result

    def make_workbook(self) -> Workbook:
        """Make workbook.

        Returns:
            Workbook.

        """
        result = Workbook()
        setattr(self, "_workbook", result)
        return result

    def save(self) -> None:
        """Save workbook."""
        workbook = self.get_workbook()
        workbook.save(self._file)


class XlsxIOWrapper(OpenPyXLMixin, AbstractIOWrapper):
    """Implements a I/O wrapper for `.xlsx` files."""

    def __init__(self, __file: IO[Any], /, mode: str) -> None:
        self._file = __file
        self._mode = mode
        self._context: Optional[AbstractXlsxWrapper] = None

    @property
    def file(self) -> IO[Any]:
        """File."""
        return self._file

    @property
    def closed(self) -> bool:
        """Whether file is closed."""
        return self._file.closed

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        default = getattr(self._context, "fieldnames", [])
        result = getattr(self, "_fieldnames", default)
        return result

    @fieldnames.setter
    def fieldnames(self, value: Any) -> None:
        if not isinstance(value, Sequence):
            message = f"expected type 'Sequence', got {type(value)} instead"
            raise TypeError(message)

        setattr(self, "_fieldnames", value)

    @property
    def sheetname(self) -> Optional[str]:
        """Sheetname."""
        result = getattr(self, "_sheetname", None)
        return result

    @sheetname.setter
    def sheetname(self, value: Any) -> None:
        if value is not None and not isinstance(value, str):
            message = f"expected type 'str', got {type(value)} instead"
            raise TypeError(message)

        setattr(self, "_sheetname", value)

    @property
    def read_only(self) -> bool:
        """Whether read only."""
        return getattr(self._context, "read_only")  # type: bool

    def __enter__(self) -> XlsxIOWrapper:
        return self

    def __exit__(self, *_: Any) -> None:
        super().__exit__(*_)
        self.close()
        return

    def close(self) -> None:
        """Close `.xlsx` file."""
        self._file.close()

    def read(self, size: int = -1, /) -> bytes:
        """Read content of `.xlsx` file.

        Returns:
            Content.

        """
        result = self._file.read(size)
        return result

    def read_record(self) -> Dict[str, Any]:
        """Read record from `.xlsx` file.

        Returns:
            Record.

        """
        reader = self._get_record_reader()
        result: Dict[str, Any] = reader.read_record()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return result

    def read_records(self) -> List[Dict[str, Any]]:
        """Read records from `.xlsx` file.

        Returns:
            Records.

        """
        reader = self._get_record_reader()
        results = reader.read_records()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return results

    def _get_record_reader(self) -> _OpenPyXLRecordReader:
        """Get `.xlsx` record reader.

        Returns:
            Reader.

        """
        if not hasattr(self, "_record_reader"):
            self._start_record_reader()

        result: _OpenPyXLRecordReader = getattr(self, "_record_reader")
        return result

    def _start_record_reader(self) -> None:
        """Start `.xlsx` record reader."""
        reader = _OpenPyXLRecordReader(self)
        setattr(self, "_record_reader", reader)

    def read_row(self) -> List[str]:
        """Read row from `.xlsx` file.

        Returns:
            Row.

        """
        reader = self._get_row_reader()
        result = reader.read_row()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return result

    def read_rows(self) -> List[List[str]]:
        """Read rows from `.xlsx` file.

        Returns:
            Rows.

        """
        reader = self._get_row_reader()
        results = reader.read_rows()

        if not self.fieldnames:
            self.fieldnames = reader.fieldnames

        return results

    def _get_row_reader(self) -> _OpenPyXLRowReader:
        """Get `.xlsx` row reader.

        Returns:
            Reader.

        """
        if not hasattr(self, "_row_reader"):
            self._start_row_reader()

        result: _OpenPyXLRowReader = getattr(self, "_row_reader")
        return result

    def _start_row_reader(self) -> None:
        """Start `.xlsx` row reader."""
        reader = _OpenPyXLRowReader(self)
        setattr(self, "_row_reader", reader)

    def write(self, buffer: memoryview, /) -> int:  # type: ignore
        """Write buffer to `.xlsx` file.

        Args:
            buffer: Buffer.

        """
        result = self._file.write(buffer)
        return result

    def write_header(self) -> None:
        """Write header."""
        writer = self._get_record_writer()
        writer.write_header()

    def write_record(self, record: Dict[str, Any], /) -> None:
        """Write record to `.xlsx` file.

        Args:
            record: Record to write.

        """
        writer = self._get_record_writer()
        writer.write_record(record)

    def write_records(self, records: Iterable[Dict[str, Any]], /) -> None:
        """Write records to `.xlsx` file.

        Args:
            records: Records to write.

        """
        writer = self._get_record_writer()
        writer.write_records(records)

    def _get_record_writer(self) -> _OpenPyXLRecordWriter:
        """Get record writer for `.xlsx` file.

        Returns:
            Writer.

        """
        if not hasattr(self, "_record_writer"):
            self._start_record_writer()

        result: _OpenPyXLRecordWriter = getattr(self, "_record_writer")
        return result

    def _start_record_writer(self) -> None:
        """Start record writer for `.xlsx` file."""
        writer = _OpenPyXLRecordWriter(self)
        setattr(self, "_record_writer", writer)

    def write_row(self, row: Iterable[Any], /) -> None:
        """Write row to `.xlsx` file.

        Args:
            row: Row to write.

        """
        writer = self._get_row_writer()
        writer.write_row(row)

    def write_rows(self, rows: Iterable[Iterable[Any]], /) -> None:
        """Write rows to `.xlsx` file.

        Args:
            rows: Rows to write.

        """
        writer = self._get_row_writer()
        writer.write_rows(rows)

    def _get_row_writer(self) -> _OpenPyXLRowWriter:
        """Get row `.xlsx` writer.

        Returns:
            Writer.

        """
        if not hasattr(self, "_writer"):
            self._start_row_writer()

        result: _OpenPyXLRowWriter = getattr(self, "_writer")
        return result

    def _start_row_writer(self) -> None:
        """Start standard `.xlsx` writer."""
        writer = _OpenPyXLRowWriter(self)
        setattr(self, "_writer", writer)


class _OpenPyXLRecordReader(Iterator[Any]):
    """Implements an OpenPyXL record reader for `.xlsx` files."""

    def __init__(self, __wrapper: "XlsxIOWrapper") -> None:
        self._worksheet = __wrapper.get_worksheet()
        self._reader = self._worksheet.iter_rows(values_only=True)
        self._fieldnames = __wrapper.fieldnames
        self._row_num = 0

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames

    @fieldnames.setter
    def fieldnames(self, fieldnames: Sequence[str]) -> None:
        self._fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._row_num

    def __next__(self) -> Dict[str, Any]:
        result = self.read_record()
        return result

    def read_record(self) -> Dict[str, Any]:
        """Read record from `.csv` file.

        Returns
            Record.

        """
        if self.row_num == 0 and self.fieldnames:
            result = self._read_first_record()
            return result

        if self.row_num == 0 and not self.fieldnames:
            self._read_header()

        result = self._read_next_record()
        return result

    def read_records(self) -> List[Dict[str, Any]]:
        """Read rows from `.xlsx` file.

        Returns:
            Rows.

        """
        if self.row_num == 0 and self.fieldnames:
            first_row = self._read_first_record()
            remainder = [self._make_record(row) for row in self._reader]
            self._row_num += len(remainder)
            results = [first_row, *remainder]
            return results

        if self.row_num == 0 and not self.fieldnames:
            self._read_header()

        results = [self._make_record(row) for row in self._reader]
        self._row_num += len(results)
        return results

    def _read_header(self) -> None:
        """Read header.

        Raises:
            RuntimeError: when `fieldnames` is not empty.
            RuntimeError: when not on first line of file.

        """
        if self.fieldnames:  # sanity check
            message = "'fieldnames' is not empty"
            raise RuntimeError(message)

        if self.row_num != 0:
            message = "reader is not on first line of file"
            raise RuntimeError(message)

        self.fieldnames = self._read_next_row()

    def _read_next_row(self) -> List[str]:
        """Read next row.

        Returns:
            Row.

        """
        row = next(self._reader)
        result = self._make_row(row)
        self._row_num += 1
        return result

    def _make_row(self, __row: Tuple[Any, ...], /) -> List[Any]:
        """Make row.

        Returns:
            Row.

        """
        result = list(__row)
        return result

    def _read_first_record(self) -> Dict[str, Any]:
        """Read first row.

        Args:
            reader: Reader.

        Raises:
            RuntimeError: when `reader` is not on first line of file.

        """
        if self.row_num != 0:
            message = "reader is not on first line of file"
            raise RuntimeError(message)

        first_record = self._read_next_record()
        if not self._is_header(list(first_record.values())):
            return first_record

        result = self._read_next_record()
        return result

    def _read_next_record(self) -> Dict[str, str]:
        """Read next record.

        Returns:
            Row.

        """
        row = next(self._reader)
        result = self._make_record(row)
        self._row_num += 1
        return result

    def _make_record(self, __row: Tuple[Any, ...], /) -> Dict[str, Any]:
        """Make record from row.

        Args:
            row: Row.

        Returns:
            Record.

        """
        result = {key: value for key, value in zip(self.fieldnames, __row)}
        return result

    def _is_header(self, row: Sequence[str]) -> bool:
        """Check whether row is header.

        Args:
            row: Row.

        Returns:
            Whether row is header.

        Raises:
            RuntimeError: when `fieldnames` is empty.

        """
        if not self.fieldnames:  # sanity check
            message = "'fieldnames' is empty"
            raise RuntimeError(message)

        result = not set(self.fieldnames).difference(row)
        return result


class _OpenPyXLRecordWriter:
    """Implements an OpenPyXL record writer for `.xlsx` files."""

    def __init__(self, __wrapper: "XlsxIOWrapper") -> None:
        self._worksheet = __wrapper.get_worksheet()
        self._fieldnames = __wrapper.fieldnames
        self._row_num = 0

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames

    @fieldnames.setter
    def fieldnames(self, fieldnames: Sequence[str]) -> None:
        self._fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._row_num

    def write_header(self) -> None:
        """Write header."""
        row = self._row_num + 1  # OpenPyXl rows are not zero-indexed
        for col, fieldname in enumerate(self.fieldnames, start=1):
            self._worksheet.cell(row, col, value=fieldname)

        self._row_num += 1

    def write_record(self, __record: Dict[str, Any], /) -> None:
        """Write record to `.xlsx` file.

        Args:
            __record: Record to write.

        """
        row = self._row_num + 1  # OpenPyXl rows are not zero-indexed
        for col, key in enumerate(self.fieldnames, start=1):
            self._worksheet.cell(row, col, value=__record[key])

        self._row_num += 1

    def write_records(self, __records: Iterable[Dict[str, Any]], /) -> None:
        """Write records to `.csv` file.

        Args:
            __records: Records to write.

        """
        start = self._row_num + 1  # OpenPyXl rows are not zero-indexed
        for row, record in enumerate(__records, start=start):
            for col, key in enumerate(self.fieldnames, start=1):
                self._worksheet.cell(row, col, value=record[key])

        self._row_num += len(list(__records))


class _OpenPyXLRowReader(Iterator[Any]):
    """Implements an OpenPyXL row reader for `.xlsx` files."""

    def __init__(self, __wrapper: "XlsxIOWrapper") -> None:
        self._worksheet = __wrapper.get_worksheet()
        self._reader = self._worksheet.iter_rows(values_only=True)
        self._fieldnames = __wrapper.fieldnames
        self._row_num = 0

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames

    @fieldnames.setter
    def fieldnames(self, fieldnames: Sequence[str]) -> None:
        self._fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._row_num

    def __next__(self) -> List[Any]:
        result = self.read_row()
        return result

    def read_row(self) -> List[str]:
        """Read row from `.csv` file.

        Returns
            Row.

        """
        if self.row_num == 0 and self.fieldnames:
            result = self._read_first_row()
            return result

        if self.row_num == 0 and not self.fieldnames:
            self._read_header()

        result = self._read_next_row()
        return result

    def read_rows(self) -> List[List[str]]:
        """Read rows from `.csv` file.

        Returns:
            Rows.

        """
        if self.row_num == 0 and self.fieldnames:
            first_row = self._read_first_row()
            remainder = [self._make_row(row) for row in self._reader]
            self._row_num += len(remainder)
            results: List[List[Any]] = [first_row, *remainder]
            return results

        if self.row_num == 0 and not self.fieldnames:
            self._read_header()

        results = [self._make_row(row) for row in self._reader]
        self._row_num += len(results)
        return results

    def _read_header(self) -> None:
        """Read header.

        Raises:
            RuntimeError: when `fieldnames` is not empty.
            RuntimeError: when not on first line of file.

        """
        if self.fieldnames:  # sanity check
            message = "'fieldnames' is not empty"
            raise RuntimeError(message)

        if self.row_num != 0:
            message = "reader is not on first line of file"
            raise RuntimeError(message)

        self.fieldnames = self._read_next_row()

    def _read_first_row(self) -> List[str]:
        """Read first row.

        Args:
            reader: Reader.

        Raises:
            RuntimeError: when `reader` is not on first line of file.

        """
        if self.row_num != 0:
            message = "reader is not on first line of file"
            raise RuntimeError(message)

        first_row = self._read_next_row()
        if not self._is_header(first_row):
            return first_row

        result = self._read_next_row()
        return result

    def _read_next_row(self) -> List[str]:
        """Reqad next row.

        Returns:
            Row.

        """
        row = next(self._reader)
        result = self._make_row(row)
        self._row_num += 1
        return result

    def _make_row(self, __row: Tuple[Any, ...], /) -> List[Any]:
        """Make row.

        Returns:
            Row.

        """
        result = list(__row)
        return result

    def _is_header(self, row: Sequence[str]) -> bool:
        """Check whether row is header.

        Args:
            row: Row.

        Returns:
            Whether row is header.

        Raises:
            RuntimeError: when `fieldnames` is empty.

        """
        if not self.fieldnames:  # sanity check
            message = "'fieldnames' is empty"
            raise RuntimeError(message)

        result = not set(self.fieldnames).difference(row)
        return result


class _OpenPyXLRowWriter:
    """Implements an OpenPyXL row writer for `.xlsx` files."""

    def __init__(self, __wrapper: "XlsxIOWrapper") -> None:
        self._worksheet = __wrapper.get_worksheet()
        self._fieldnames = __wrapper.fieldnames
        self._row_num = 0

    @property
    def fieldnames(self) -> Sequence[str]:
        """Fieldnames."""
        return self._fieldnames

    @fieldnames.setter
    def fieldnames(self, fieldnames: Sequence[str]) -> None:
        self._fieldnames = fieldnames

    @property
    def row_num(self) -> int:
        """Row number."""
        return self._row_num

    def write_header(self) -> None:
        """Write header."""
        self.write_row(self.fieldnames)

    def write_row(self, __row: Iterable[Any], /) -> None:
        """Write row to `.xlsx` file.

        Args:
            row: Row to write.

        """
        row = self._row_num + 1  # OpenPyXl rows are not zero-indexed
        for col, fieldname in enumerate(__row, start=1):
            self._worksheet.cell(row, col, value=fieldname)

        self._row_num += 1

    def write_rows(self, __rows: Iterable[Iterable[Any]], /) -> None:
        """Write rows to `.xlsx` file.

        Args:
            __rows: Rows to write.

        """
        start = self._row_num + 1  # OpenPyXl rows are not zero-indexed
        for row, data in enumerate(__rows, start=start):
            for col, value in enumerate(data, start=1):
                self._worksheet.cell(row, col, value=value)

        self._row_num += len(list(__rows))



# -*- coding: utf-8 -*-
"""Message class."""

# pylint: disable=too-few-public-methods

# Standard Library Imports
from __future__ import annotations
import abc
import datetime

# Local Imports
from ..metaclasses import MessageMeta

__all__ = ["AbstractMessage"]


class AbstractMessage(abc.ABC, metaclass=MessageMeta):
    """Class represents an abstract message."""

    __created_at__: datetime.datetime

    def __gt__(self, other: object) -> bool:
        result = (
            self.__created_at__ > other.__created_at__
            if isinstance(other, AbstractMessage)
            else False
        )
        return result

    def __lt__(self, other: object) -> bool:
        result = (
            self.__created_at__ < other.__created_at__
            if isinstance(other, AbstractMessage)
            else False
        )
        return result

    def __repr__(self) -> str:
        result = self.__class__.__name__
        return result

    def __str__(self) -> str:
        result = self.__class__.__name__
        return result



# -*- coding: utf-8 -*-
"""Class Mapper."""

# Standard Library Imports
from contextlib import contextmanager
import typing

# Local Imports
from .abstract_mapper import AbstractMapper

__all__ = ["ClassMapper"]


# Custom types
T = typing.TypeVar("T")


class ClassMapper(AbstractMapper):
    """Implements a class mapper."""

    def __init__(
        self,
        __class: typing.Type[T],
        /,
        schema: typing.Dict[str, typing.Any],
        properties: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> None:
        super().__init__(schema)
        self._class = __class
        self._properties = properties

        setattr(self._class, "__mapper__", self)

    @property
    def cls(self) -> typing.Type[T]:
        """Mapped class."""
        return self._class

    @property
    def properties(self) -> typing.Dict[str, typing.Any]:
        """Mapped properties."""
        results = self._properties or self._get_attribute_names()
        return results

    def from_dict(self, __dict: dict, /) -> T:
        """Instantiate class from dictionary.

        Args:
            __dict: Dictionary.

        Returns:
            Instance of mapped class.

        """
        with replace_init(self.cls) as cls:
            result = cls()
            for attr, key in self.properties.items():
                converter = self._get_converter(key, direction="outward")
                value = converter(__dict[key]) if converter else __dict[key]
                setattr(result, attr, value)

        return result

    def from_list(self, __list: list, /) -> T:
        """Instantiate class from list.

        Args:
            __list: List.

        Returns:
            Instance of mapped class.

        """
        with replace_init(self.cls) as cls:
            result = cls()
            for attr, idx in self.properties.items():
                converter = self._get_converter(idx, direction="outward")
                value = converter(__list[idx]) if converter else __list[idx]
                setattr(result, attr, value)

        return result

    def to_dict(self, __instance: T, /) -> dict:
        """Convert instance of class to dictionary.

        Args:
            __instance: Instance of mapped class.

        Returns:
            Dictionary.

        """
        result = {}
        for attr, key in self.properties.items():
            converter = self._get_converter(key, direction="inward")
            value = (
                converter(getattr(__instance, attr))
                if converter
                else getattr(__instance, attr)
            )
            result[key] = value

        return result

    def to_list(self, __instance: T, /) -> list:
        """Convert instance of class to list.

        Args:
            __instance: Instance of mapped class.

        Returns:
            List.

        """
        result = []
        for attr, key in sorted(self.properties.items(), key=lambda i: i[1]):
            converter = self._get_converter(key, direction="inward")
            value = (
                converter(getattr(__instance, attr))
                if converter
                else getattr(__instance, attr)
            )
            result.append(value)

        return result


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
@contextmanager
def replace_init(__class: type) -> typing.Generator[type, None, None]:
    """Temporarily replace the ``__init__`` method on a class.

    Args:
        __class: Class.

    Yields:
        Class.

    """
    init = getattr(__class, "__init__")
    repl = getattr(object, "__init__")

    setattr(__class, "__init__", repl)
    yield __class
    setattr(__class, "__init__", init)



# -*- coding: utf-8 -*-
"""Tqdm Progress Bar."""

# Standard Library Imports
from typing import Any
from typing import Optional
from typing import Union

# Third-Party Imports
from tqdm import tqdm

# Local Imports
from .abstract_progress_bar import AbstractProgressBar

__all__ = ["TqdmProgressBar"]


class TqdmProgressBar(AbstractProgressBar):
    """Implements a tqdm progress bar."""

    def __init__(
        self,
        desc: Optional[str] = None,
        total: Optional[Union[float, int]] = None,
        leave: bool = False,
        miniters: Union[float, int] = 1,
        **kwargs: Any,
    ) -> None:
        self._progress_bar = tqdm(
            desc=desc,
            total=total,
            leave=leave,
            miniters=miniters,
            **kwargs,
        )

    @property
    def current(self) -> int:
        """Current progress."""
        return self._progress_bar.n

    @property
    def leave(self) -> bool:
        """Whether progress bar is maintained between iterations."""
        return self._progress_bar.leave

    @leave.setter
    def leave(self, value: bool) -> None:
        if not isinstance(value, bool):  # type: ignore
            message = f"expected type 'bool', got {type(value)} instead"
            raise TypeError(message)

        self._progress_bar.leave = value

    @property
    def total(self) -> int:
        """Total progress."""
        return self._progress_bar.total

    @total.setter
    def total(self, n: Any) -> None:
        if not isinstance(n, (float, int)):
            expected = "expected type 'float' or 'int'"
            actual = f"got {n} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        self._progress_bar.total = n

    def close(self) -> None:
        """Close progress bar."""
        self._progress_bar.close()

    def refresh(self) -> None:
        """Refresh progress bar."""
        self._progress_bar.refresh()

    def reset(self, total: Optional[Union[float, int]] = None) -> None:
        """Reset progress bar."""
        self._progress_bar.reset(total)

    def update(self, n: Union[float, int] = 1) -> None:
        """Update progress bar.

        Args:
            n (optional): Amount by which to increment the internal counter. Defaul ``1``.

        """
        self._progress_bar.update(n)

    def write(self, message: str) -> None:
        """Write message to progress bar."""
        self._progress_bar.write(message)



# -*- coding: utf-8 -*-
"""Currency Converter.

Module provides function for converting values to currencies.

"""

# Standard Library Imports
import decimal
import re
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_currency"]


def to_currency(__value: Any, /, default: float = 0.00) -> float:
    """Convert value to currency.

    Args:
        __value: Value to convert to currency.
        default (optional): Default value. Default ``0.00``.

    Returns:
        Amount.

    """
    converter = CurrencyConverter(default=default)
    result = converter(__value)
    return result


class CurrencyConverter(BaseConverter):
    """Class implements a currency converter.

    Args:
        default (optional): Default value. Default ``0.00``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: float = 0.00,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if not isinstance(default, float):
            message = f"expected type 'float', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()


def currency_from_decimal(__value: decimal.Decimal, _: float, /) -> float:
    """Convert decimal value to currency.

    Args:
        __value: Value to convert to currency.

    Returns:
        Amount.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = float(round(__value, 2))
    return result


def currency_from_float(__value: float, _: float, /) -> float:
    """Convert float value to currency.

    Args:
        __value: Value to convert to currency.

    Returns:
        Amount.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = round(__value, 2)
    return result


def currency_from_int(__value: int, _: float, /) -> float:
    """Convert integer value to currency.

    Args:
        __value: Value to convert to currency.

    Returns:
        Amount.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = round(float(__value), 2)
    return result


def currency_from_str(__value: str, default: float = 0.00, /) -> float:
    """Convert string value to currency.

    Args:
        __value: String representation of currency value.
        default (optional): Default value. Default ``0.00``.

    Returns:
        Amount.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to currency.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    if not value:
        return default

    try:
        amount = float(re.sub(r"[^0-9a-zA-Z.]+", r"", value))

    except ValueError:
        message = f"'{__value}' cannot be converted to currency"
        raise ValueError(message)

    result = round(amount, 2)
    return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., float]] = {
    decimal.Decimal: currency_from_decimal,
    float: currency_from_float,
    int: currency_from_int,
    str: currency_from_str,
}



# -*- coding: utf-8 -*-
"""Command class.

This module defines command classes. Commands are simple data structures that
work as messages in the system, transmitting instructions from one part of the
system to another. Commands encapsulate the intent to perform a specific
action, along with an expectation of a particular result. Commands are always
named using an imperative mood verb phrase.

Actions implemented by these commands should be restricted to write methods:
insert, update, and delete. The only time select statements should arise are
when information is required during the process of writing a record to the
database, such as retrieving the necessary database IDs for foreign keys.

"""

# Local Imports
from .message import AbstractMessage

__all__ = ["BaseCommand"]


class BaseCommand(AbstractMessage):
    """Class implements a command."""



# -*- coding: utf-8 -*-
"""Base Converter.

Module defines a base class for converting values between data types.

"""

# Standard Library Imports
from collections import ChainMap
from typing import Any
from typing import Callable
from typing import Literal
from typing import Type
from typing import TypeVar

# Local Imports
from .abstract_converter import AbstractConverter

__all__ = ["BaseConverter"]


# Custom type
T = TypeVar("T")


class BaseConverter(AbstractConverter):
    """Implements a base class for converters.

    Args:
        default (optional): Default value. Default ``None``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    _conversions: ChainMap[Type[Any], Callable[..., Any]]

    def __init__(
        self,
        *,
        default: Any = None,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        self._conversions = ChainMap()
        self._default = default

        if on_error not in ("default", "raise"):
            message = "invalid value for 'on_error' argument"
            raise ValueError(message)

        self._on_error = on_error

    def __call__(self, __value: Any, /) -> Any:
        try:
            result = (
                self._handle_conversion(__value)
                if __value is not None
                else self._default
            )
        except KeyError as error:
            result = self._handle_exception(__value, error)

        return result

    def _handle_conversion(self, __value: Any, /) -> Any:
        """Handle conversion between data types.

        Args:
            __value: Value to convert.

        Returns:
            Converted value.

        """
        conversion = self._conversions[type(__value)]
        result = conversion(__value, self._default)
        return result

    def _handle_exception(self, __value: Any, error: Exception, /) -> Any:
        """Handle conversion between data types.

        Args:
            __value: Value to convert.

        Returns:
            Converted value.

        """
        if self._on_error == "raise":
            message = f"unable to convert from {type(__value)}"
            raise TypeError(message) from error

        return self._default

    def set_conversion(
        self, __type: type, func: Callable[..., Any], /
    ) -> None:
        """Set conversion for type.

        Args:
            __type: Type.
            func: Function for converting between types.

        """
        self._conversions[__type] = func

    def reset(self) -> None:
        """Reset converter."""
        self._conversions = (
            self._conversions.parents
            if len(self._conversions.maps) > 1
            else self._conversions
        )



# -*- coding: utf-8 -*-
"""Date Converter.

Module provides function for converting values to dates.

"""

# Standard Library Imports
import datetime
import decimal
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional
from typing import Union

# Third-Party Imports
import cachetools
from cachetools.keys import hashkey
from dateutil.parser import parse
from dateutil.parser import ParserError

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_date"]


def to_date(
    __value: Any, /, default: Optional[datetime.date] = None
) -> Optional[datetime.date]:
    """Converts value to date.

    Args:
        __value: Date.
        default (optional): Default value. Default ``None``.

    Returns:
        Date.

    """
    converter = DateConverter(default=default)
    result = converter(__value)
    return result


class DateConverter(BaseConverter):
    """Class implements a date converter.

    Args:
        default (optional): Default value. Default ``None``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: Optional[datetime.date] = None,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if default and not isinstance(default, datetime.date):  # type: ignore
            message = f"expected type 'date', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()


def date_from_date(
    __value: datetime.date, _: Optional[datetime.date], /
) -> datetime.date:
    """Convert date value to ``date``.

    Args:
        __value: Value to convert to ``date``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    result = datetime.date(__value.year, __value.month, __value.day)
    return result


def date_from_datetime(
    __value: datetime.datetime, _: Optional[datetime.date], /
) -> datetime.date:
    """Convert datetime value to ``date``.

    Args:
        __value: Value to convert to ``date``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    result = __value.date()
    return result


def date_from_decimal(
    __value: decimal.Decimal, _: Optional[datetime.date], /
) -> datetime.date:
    """Convert decimal value to ``date``.

    Args:
        __value: Value to convert to ``date``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = date_from_float(float(__value), _)
    return result


def date_from_float(
    __value: float, _: Optional[datetime.date], /
) -> datetime.date:
    """Convert float value to ``date``.

    Args:
        __value: Value to convert to ``date``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    try:
        result = date_from_serial_date(__value, _)

    except ValueError:
        result = date_from_timestamp(__value, _)

    return result


def date_from_serial_date(
    __value: float, _: Optional[datetime.date], /
) -> datetime.date:
    """Convert serial date value to ``date``.

    Implementation based on answer from Stack Overflow.

    Args:
        __value: Value to convert to ``date``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'float'.

    .. _Stack Overflow:
        https://stackoverflow.com/a/6706556/16732779

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    seconds = (__value - 25569) * 86400.0
    result = date_from_timestamp(seconds, _)
    return result


def date_from_int(
    __value: int, _: Optional[datetime.date], /
) -> datetime.date:
    """Convert integer value to ``date``.

    Args:
        __value: Value to convert to ``date``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = date_from_timestamp(__value, _)
    return result


@cachetools.cached(cachetools.LRUCache(maxsize=1000), hashkey)
def date_from_str(
    __value: str, default: Optional[datetime.date] = None, /
) -> Optional[datetime.date]:
    """Convert string value to ``date``.

    Args:
        __value: String representation of date.
        default (optional): Default value. Default ``None``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to ``date``.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    try:
        value = __value.replace("  ", " ").strip()
        result = parse(value).date() if value else default

    except (ParserError, ValueError) as error:
        message = f"'{__value}' cannot be converted to date"
        raise ValueError(message) from error

    return result


def date_from_timestamp(
    __value: Union[float, int], _: Optional[datetime.date], /
) -> datetime.date:
    """Convert timestamp value to ``date``.

    Args:
        __value: Value to convert to ``date``.

    Returns:
        Date.

    Raises:
        TypeError: when value is not type 'float' or 'int'.

    """
    if not isinstance(__value, (float, int)):  # type: ignore
        expected = "expected type 'float' or 'int'"
        actual = f"got {type(__value)} instead"
        message = ", ".join([expected, actual])
        raise TypeError(message)

    result = datetime.datetime.fromtimestamp(__value).date()
    return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., Optional[datetime.date]]] = {
    datetime.date: date_from_date,
    datetime.datetime: date_from_datetime,
    decimal.Decimal: date_from_decimal,
    float: date_from_float,
    int: date_from_int,
    str: date_from_str,
}



# -*- coding: utf-8 -*-
"""Sessioned Unit of Work.

Based on 'Architecture Patterns in Python' unit-of-work pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# Standard Library Imports
from __future__ import annotations
from typing import Any
from typing import Callable
from typing import Optional
from typing import TypeVar

# Local Imports
from .abstract_unit_of_work import AbstractUnitOfWork


# Custom types
T = TypeVar("T")

# Private attributes
SESSION_ATTR = "_session"


class SessionedUnitOfWork(AbstractUnitOfWork):
    """Class implements a sessioned unit of work.

    Args:
        *args (optional): Positional arguments.
        session_factory (optional): Function for creating a session.
        **kwargs (optional): Keyword arguments.

    Attributes:
        session: Session.

    """

    def __init__(
        self,
        *args: Any,
        session_factory: Callable[..., T],
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)
        self._session_factory = session_factory

    @property
    def session(self) -> Optional[T]:  # type: ignore
        """Session."""
        return getattr(self, SESSION_ATTR, None)

    def __enter__(self) -> SessionedUnitOfWork:
        setattr(self, SESSION_ATTR, self._session_factory())
        super().__enter__()
        return self

    def __exit__(self, *args: Any) -> None:
        super().__exit__(*args)
        self.close()

    def close(self) -> None:
        """Close session."""
        method = getattr(self.session, "close", None)
        if method is not None:
            method()

    def commit(self) -> None:
        """Commit changes."""
        super().commit()  # type: ignore
        method = getattr(self.session, "commit", None)
        if method is not None:
            method()

    def rollback(self) -> None:
        """Rollback changes."""
        super().rollback()  # type: ignore

        method = getattr(self.session, "rollback", None)
        if method is not None:
            method()



# -*- coding: utf-8 -*-
"""Txt Repository."""

# Standard Library Imports
import logging
from typing import Any

# Local Imports
from .file_system_repository import AbstractFileSystemRepository
from ..wrappers.txt_file_wrappers import AbstractTxtWrapper

__all__ = ["AbstractTxtRepository"]


# Initiate logger.
log = logging.getLogger("dodecahedron")


class AbstractTxtRepository(AbstractFileSystemRepository):
    """Represents an abstract `.txt` file repository.

    Args:
        __file: Text file.

    """

    def __init__(self, __file: Any, /, *args: Any, **kwargs: Any) -> None:
        if not isinstance(__file, AbstractTxtWrapper):
            expected = "expected type 'AbstractTxtWrapper'"
            actual = f"got {type(__file)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        super().__init__(__file, *args, **kwargs)



# -*- coding: utf-8 -*-
"""Abstract Wrapper."""

# Standard Library Imports
import abc
import logging
import os
import pathlib
from typing import Any
from typing import IO
from typing import Optional
from typing import TypeVar

# Local Imports
from .. import settings
from .. import utils

__all__ = [
    "AbstractFileSystemWrapper",
    "AbstractDirectoryWrapper",
    "AbstractFileWrapper",
]

# Initiate logger.
log = logging.getLogger("dodecahedron")

# Custom types
T = TypeVar("T")


class AbstractFileSystemWrapper(abc.ABC):
    """Represents an abstract file-system wrapper."""

    def __init__(self, *, read_only: bool = False) -> None:
        self.read_only = read_only

    @property
    @abc.abstractmethod
    def extension(self) -> str:
        """File extension."""
        raise NotImplementedError

    @property
    def read_only(self) -> bool:
        """Whether read only."""
        return self._read_only

    @read_only.setter
    def read_only(self, value: bool) -> None:
        if not isinstance(value, bool):  # type: ignore
            message = f"expected type 'bool', got {type(value)} instead"
            raise TypeError(message)

        self._read_only = value

    @abc.abstractmethod
    def open(self, *args: Any, **kwargs: Any) -> "IO[Any]":
        """Open file."""
        raise NotImplementedError

    def _open_file(
        self,
        filepath: "os.PathLike[Any]",
        mode: str,
    ) -> "IO[Any]":
        """Open a file and return a file object.

        Args:
            filepath: Path to file.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        """
        self._raise_for_mode(mode)
        result = open(filepath, mode=mode)
        return result

    def _raise_for_mode(self, mode: str) -> None:
        """Raise for invalid mode.

        Args:
            mode: Mode.

        Raises:
            ValueError: when mode is invalid.

        """
        if self.read_only and set(mode).intersection({"a", "w", "x"}):
            message = f"{mode!s} mode not allowed when read-only is 'True'"
            raise ValueError(message)


class AbstractDirectoryWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for directories.

    Args:
        directory: Directory which contains file(s).
        read_only (optional): Whether directory is read only. Default ``False``.

    Raises:
        TypeError: when `directory` is not type ``PathLike``.
        NotADirectoryError: when `directory` is not a valid directory.

    """

    def __init__(
        self,
        directory: "os.PathLike[Any]",
        *,
        extension: Optional[str] = None,
        read_only: bool = False,
    ) -> None:
        if not isinstance(directory, os.PathLike):  # type: ignore
            expected = "expected type 'PathLike'"
            actual = f"got {type(directory)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        if extension and not isinstance(extension, str):  # type: ignore
            message = f"expected type 'str', got {type(extension)} instead"
            raise TypeError(message)

        if not os.path.exists(directory) or not os.path.isdir(directory):
            message = f"{directory!s} is not a valid directory"
            raise NotADirectoryError(message)

        super().__init__(read_only=read_only)
        self._directory = pathlib.Path(directory)
        self._extension = utils.standardize_file_extension(extension or "*")

    @property
    def directory(self) -> pathlib.Path:
        """Path to directory."""
        return self._directory

    @property
    def extension(self) -> str:
        """File extension."""
        return self._extension

    def __fspath__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return str(self.directory)

    def open(self, filename: str, /, mode: str = "r") -> "IO[Any]":
        """Open a file and return a file object.

        Args:
            filename: Filename.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        Raises:
            TypeError: when `encoding` is not type ``str``.

        """
        filepath = self._get_filepath(filename)
        result = self._open_file(filepath, mode)
        return result

    def _get_filepath(self, filename: str) -> pathlib.Path:
        """Get filepath.

        Args:
            filename: Filename.

        Returns:
            Filepath.

        """
        result = self._directory / filename
        if not result.exists():
            result = self.find(filename)

        utils.raise_for_extension(result, self.extension)
        return result

    def find(self, ref: str) -> pathlib.Path:
        """Find path for file in directory.

        Finds the filepath for a file in the source directory where the
        filename contains the provided substring.

        Args:
            ref: Substring for which to search.

        Returns:
            Path for file.

        Raises:
            FileNotFoundError: When no filenames match provided substring.

        """
        log.debug(
            "Searching for %(ref)s in %(dir)s",
            {"ref": ref, "dir": self._directory},
        )
        filename = utils.set_extension(ref, self.extension)
        result = self._find_file_in_directory(filename)
        log.debug(
            "Found %(file)s in %(dir)s",
            {"file": result, "dir": self._directory},
        )
        return result

    def _find_file_in_directory(self, __filename: str) -> pathlib.Path:
        """Find path for file in directory.

        Finds the filepath for a file in the source directory where the
        filename contains the provided substring.

        Args:
            __filename: Name of file.

        Returns:
            Path for file.

        Raises:
            FileNotFoundError: When no files match provided filename.

        """
        try:
            filepath = next(path for path in self._directory.rglob(__filename))

        except StopIteration as err:
            message = f"{self._directory / __filename} not found"
            raise FileNotFoundError(message) from err

        return filepath


class AbstractFileWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for files.

    Args:
        filepath: Path to file.
        read_only (optional): Whether file is read only. Default ``False``.

    Raises:
        TypeError: when `filepath` is not type ``PathLike``.
        IsADirectoryError: when `filepath` points to a directory.

    """

    def __init__(
        self,
        filepath: "os.PathLike[Any]",
        *,
        read_only: bool = False,
    ) -> None:
        if not isinstance(filepath, os.PathLike):  # type: ignore
            expected = "expected type 'PathLike'"
            actual = f"got {type(filepath)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        if os.path.isdir(filepath):
            message = f"{filepath!s} is a directory"
            raise IsADirectoryError(message)

        super().__init__(read_only=read_only)
        self._filepath = pathlib.Path(filepath)

    @property
    def filepath(self) -> pathlib.Path:
        """Path to file."""
        return self._filepath

    @property
    def extension(self) -> str:
        """File extension."""
        return self._filepath.suffix

    def __fspath__(self) -> str:
        return str(self)

    def __str__(self) -> str:
        return str(self.filepath)

    def open(self, mode: str = "rb") -> "IO[Any]":
        """Open file and return a file object.

        Args:
            mode (optional): Mode. Default ``rb``.

        Returns:
            File object.

        """
        result = self._open_file(self._filepath, mode=mode)
        return result


class AbstractTextWrapper(AbstractFileSystemWrapper):
    """Represents an abstract wrapper class for text files.

    Args:
        *args: Positional arguments.
        encoding (optional): File encoding. Default `utf-8`.
        newline (optional): Newline character. Default ``None``.
        **kargs: Keyword arguments.

    Raises:
        TypeError: when `encoding` is not type ``str``.

    """

    def __init__(
        self,
        *args: Any,
        encoding: str = settings.DEFAULT_FILE_ENCODING,
        newline: Optional[str] = None,
        **kwargs: Any,
    ) -> None:
        super().__init__(*args, **kwargs)

        if not isinstance(encoding, str):  # type: ignore
            message = f"expected type 'str', got {type(encoding)} instead"
            raise TypeError(message)

        self._encoding = encoding
        self._newline = newline

    @property
    def encoding(self) -> str:
        """Expected file encoding."""
        return self._encoding

    @property
    def newline(self) -> Optional[str]:
        """Newline character."""
        return self._newline

    def _open_file(
        self,
        filepath: "os.PathLike[Any]",
        mode: str,
    ) -> IO[Any]:
        """Open a file and return a file object.

        Args:
            filepath: Path to file.
            mode (optional): Mode. Default ``r``.

        Returns:
            File object.

        """
        encoding = self._get_encoding_for_file_mode(mode)
        newline = self._get_newline_for_file_mode(mode)
        result = open(
            filepath,
            mode=mode,
            encoding=encoding,
            newline=newline,
        )
        return result

    def _get_encoding_for_file_mode(self, mode: str) -> Optional[str]:
        """Get encoding for file mode.

        Args:
            mode: Mode.

        Returns:
            Encoding.

        """
        result = self._encoding if "b" not in mode else None
        return result

    def _get_newline_for_file_mode(self, mode: str) -> Optional[str]:
        """Get newline for file mode.

        Args:
            mode: Mode.

        Returns:
            Newline.

        """
        result = self._newline if "b" not in mode else None
        return result


class AbstractIOWrapper(IO[Any]):  # type: ignore
    """Represents an abstract I/O wrapper class for files.

    Args:
        __file: File.

    """

    @property
    @abc.abstractmethod
    def file(self) -> IO[Any]:
        """File."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def closed(self) -> bool:
        """Whether file is closed."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def read_only(self) -> bool:
        """Whether read only."""
        raise NotImplementedError

    @abc.abstractmethod
    def close(self) -> None:
        """Close `.xlsx` file."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""Eventful Repository.

Implementation based on 'Architecture Patterns in Python' repository pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# Standard Library Imports
from collections import deque
from typing import Any
from typing import Deque
from typing import Generator
from typing import Iterable
from typing import List
from typing import Union

# Local Imports
from .abstract_repository import AbstractRepository
from ..messages import AbstractMessage
from ..messages import BaseEvent
from ..queues import MessageQueue

__all__ = ["EventfulRepository"]


class EventfulRepository(AbstractRepository):
    """Represents an eventful repository.

    Attributes:
        events: Events.

    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self._events = MessageQueue()

    @property
    def events(self) -> MessageQueue:
        """Events."""
        return self._events

    def collect_events(self) -> Generator[AbstractMessage, None, None]:
        """Collect events.

        Yields:
            Events.

        """
        self._update_events()
        while self.events:
            yield self.events.popleft()

    def _update_events(self) -> None:
        """Update events."""
        events = self._get_child_events()
        self.events.extend(events)
        self.events.sort()

    def _get_child_events(self) -> List[AbstractMessage]:
        """Get events from child objects.

        Returns:
            Events.

        """
        results = collect_events_from_objects(self.seen)
        return results


# ----------------------------------------------------------------------------
# Helper Functions
# ----------------------------------------------------------------------------
def collect_events_from_objects(
    objs: Iterable[AbstractMessage],
) -> List[AbstractMessage]:
    """Collect events from objects.

    Args:
        objs: Objects from which to collect events.

    Returns:
        Events.

    """
    results: List[AbstractMessage] = []
    for obj in objs:
        events = collect_events_from_object(obj)
        results.extend(events)

    return results


def collect_events_from_object(obj: object) -> List[AbstractMessage]:
    """Collect events from object.

    Args:
        obj: Object from which to collect events.

    Returns:
        Events.

    """
    events = get_events(obj)

    results: List[AbstractMessage] = []
    while events:
        event: AbstractMessage = events.popleft()
        results.append(event)

    return results


def get_events(obj: object) -> Union[Deque[BaseEvent], MessageQueue]:
    """Get events from object.

    Args:
        obj: Object.

    Returns:
        Events.

    """
    result = getattr(obj, "events", MessageQueue())
    if not is_message_queue(result) and is_iterable(result):
        return MessageQueue(result)

    if not is_message_queue(result) and not is_iterable(result):
        expected = "expected type 'Deque' or 'MessageQueue'"
        actual = f"got {type(result)} instead"
        message = ", ".join([expected, actual])
        raise TypeError(message)

    return result


# ----------------------------------------------------------------------------
# Validators
# ----------------------------------------------------------------------------
def is_iterable(obj: object) -> bool:
    """Check whether object is iterable.

    Args:
        obj: Object to check.

    Returns:
        Whether object is iterable.

    """
    result = hasattr(obj, "__iter__")
    return result


def is_message_queue(obj: object) -> bool:
    """Check whether object is a message queue.

    Args:
        obj: Object to check.

    Returns:
        Whether object is a queue.

    """
    result = isinstance(obj, (MessageQueue, deque))
    return result



# -*- coding: utf-8 -*-
"""Decimal Converter.

Module provides function for converting values to decimals.

"""

# Standard Library Imports
import datetime
import decimal
import re
import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal
from typing import Optional

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_decimal"]


def to_decimal(
    __value: Any, /, default: decimal.Decimal = decimal.Decimal("0.0")
) -> decimal.Decimal:
    """Convert value to decimal.

    Args:
        __value: Value to convert to decimal.
        default (optional): Default value. Default ``0.0``.

    Returns:
        Decimal.

    """
    converter = DecimalConverter(default=default)
    result = converter(__value)
    return result


class DecimalConverter(BaseConverter):
    """Class implements a decimal converter.

    Args:
        default (optional): Default value. Default ``0.0``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: decimal.Decimal = decimal.Decimal("0.0"),
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if not isinstance(default, decimal.Decimal):  # type: ignore
            message = f"expected type 'Decimal', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()


def decimal_from_bool(__value: bool, _: decimal.Decimal, /) -> decimal.Decimal:
    """Convert boolean value to decimal.

    Args:
        __value: Value to convert to decimal.

    Returns:
        Decimal.

    Raises:
        TypeError: when value is not type 'bool'.

    """
    if not isinstance(__value, bool):  # type: ignore
        message = f"expected type 'bool', got {type(__value)} instead"
        raise TypeError(message)

    result = decimal.Decimal(__value)
    return result


def decimal_from_date(
    __value: datetime.date, _: decimal.Decimal, /
) -> decimal.Decimal:
    """Convert date value to decimal.

    Args:
        __value: Value to convert to decimal.

    Returns:
        Decimal.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    value = time.mktime(__value.timetuple())
    result = decimal.Decimal(str(value))
    return result


def decimal_from_datetime(
    __value: datetime.datetime, _: decimal.Decimal, /
) -> decimal.Decimal:
    """Convert datetime value to decimal.

    Args:
        __value: Value to convert to decimal.

    Returns:
        Decimal.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.timestamp()
    result = decimal.Decimal(str(value))
    return result


def decimal_from_decimal(
    __value: decimal.Decimal, _: decimal.Decimal, /
) -> decimal.Decimal:
    """Convert decimal value to decimal.

    Args:
        __value: Value to convert to decimal.

    Returns:
        Decimal.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = decimal.Decimal(__value)
    return result


def decimal_from_float(
    __value: float, _: decimal.Decimal, /
) -> decimal.Decimal:
    """Convert float value to decimal.

    Args:
        __value: Value to convert to decimal.

    Returns:
        Decimal.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = decimal.Decimal(str(__value))
    return result


def decimal_from_int(__value: int, _: decimal.Decimal, /) -> decimal.Decimal:
    """Convert integer value to decimal.

    Args:
        __value: Value to convert to decimal.

    Returns:
        Decimal.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = decimal.Decimal(__value)
    return result


def decimal_from_str(
    __value: str,
    default: Optional[decimal.Decimal] = decimal.Decimal("0.0"),
    /,
) -> Optional[decimal.Decimal]:
    """Convert string value to decimal.

    Args:
        __value: String representation of decimal value.
        default (optional): Default value. Default ``0.0``.

    Returns:
        Decimal.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to decimal.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    if not value:
        return default

    try:
        result = decimal.Decimal(re.sub(r"[^0-9a-zA-Z.]+", r"", value))
    except decimal.InvalidOperation:
        message = f"'{__value}' cannot be converted to decimal"
        raise ValueError(message)

    return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., Optional[decimal.Decimal]]] = {
    bool: decimal_from_bool,
    datetime.date: decimal_from_date,
    datetime.datetime: decimal_from_datetime,
    decimal.Decimal: decimal_from_decimal,
    float: decimal_from_float,
    int: decimal_from_int,
    str: decimal_from_str,
}



# -*- coding: utf-8 -*-

# Standard Library Imports
import os
from typing import Optional

__all__ = [
    "is_development_environment",
    "is_production_environment",
    "is_staging_environment",
    "is_test_environment",
    "get_environment",
]


def is_development_environment() -> bool:
    """Check whether running in development environment.

    Returns:
        Whether running in development environment.

    """
    environment = get_environment()
    result = environment in ("dev", "development")
    return result


def is_production_environment() -> bool:
    """Check whether running in production environment.

    Returns:
        Whether running in production environment.

    """
    environment = get_environment()
    result = environment in ("prod", "production")
    return result


def is_staging_environment() -> bool:
    """Check whether running in staging environment.

    Returns:
        Whether running in staging environment.

    """
    environment = get_environment()
    result = environment == "staging"
    return result


def is_test_environment() -> bool:
    """Check whether running in test environment.

    Returns:
        Whether running in test environment.

    """
    environment = get_environment()
    result = environment == "test"
    return result


def get_environment() -> Optional[str]:
    """Get name of environment.

    Returns:
        Name of environment

    """
    value = os.environ.get("ENV")
    result = value.lower() if isinstance(value, str) else None
    return result



# -*- coding: utf-8 -*-
"""File Extension Utility Functions."""

# Standard Library Imports
from os import PathLike
from pathlib import PurePath
from typing import Any
from typing import TypeVar
from typing import Union

__all__ = [
    "has_extension",
    "raise_for_extension",
    "set_extension",
    "standardize_file_extension",
]


# Custom types
T = TypeVar("T")


def raise_for_extension(
    __file: Union["PathLike[Any]", PurePath, str],
    /,
    extension: str,
) -> None:
    """Raise for incorrect file extension.

    Args:
        __file: Filename or filepath.
        extension: Expected file extension.

    Raises:
        ValueError: when filepath does not have expected extension.

    """
    if not has_extension(__file, extension):
        filename = PurePath(__file).name
        message = f"{filename!s} is not a '{extension!s}' file"
        raise ValueError(message)


def has_extension(
    __file: Union["PathLike[Any]", PurePath, str],
    /,
    extension: str,
) -> bool:
    """Check whether file has expected extension.

    Args:
        __file: Filename or filepath.
        extension: Expected file extension.

    Returns:
        Whether file has expected extension.

    Raises:
        TypeError: when `filepath` is not type `Path` or `str`.
        TypeError: when `extension` is not type `str`.

    """
    if not isinstance(__file, (PathLike, PurePath, str)):  # type: ignore
        expected = "expected type 'PathLike', 'Path' or 'str'"
        actual = f"got {type(__file)} instead"
        message = ", ".join([expected, actual])
        raise TypeError(message)

    if isinstance(__file, PurePath):
        return filepath_has_extension(__file, extension)

    if isinstance(__file, str):  # type: ignore
        return filename_has_extension(__file, extension)

    return False


def filename_has_extension(__filename: str, /, extension: str) -> bool:
    """Check whether filename has expected extension.

    Args:
        __filename: Name of file.
        extension: Expected file extension.

    Returns:
        Whether file has expected extension.

    Raises:
        TypeError: when `filename` is not type `str`.
        TypeError: when `extension` is not type `str`.

    """
    if not isinstance(__filename, str):  # type: ignore
        message = f"expected type 'str', got {type(__filename)} instead"
        raise TypeError(message)

    if not isinstance(extension, str):  # type: ignore
        message = f"expected type 'str', got {type(extension)} instead"
        raise TypeError(message)

    expected = standardize_file_extension(extension)
    result = expected == "*" or __filename.endswith(f".{expected}")
    return result


def filepath_has_extension(__filepath: PurePath, /, extension: str) -> bool:
    """Check whether filepath has expected extension.

    Args:
        __filepath: Path to file.
        extension: Expected file extension.

    Returns:
        Whether file has expected extension.

    Raises:
        TypeError: when `filepath` is not type `Path`.
        TypeError: when `extension` is not type `str`.

    """
    if not isinstance(__filepath, PurePath):  # type: ignore
        message = f"expected type 'Path', got {type(__filepath)} instead"
        raise TypeError(message)

    if not isinstance(extension, str):  # type: ignore
        message = f"expected type 'str', got {type(extension)} instead"
        raise TypeError(message)

    actual = standardize_file_extension(__filepath.suffix)
    expected = standardize_file_extension(extension)
    result = expected == "*" or actual == expected
    return result


def set_extension(__file: T, extension: str) -> T:
    """Set extension on file.

    Args:
        __file: Filename or filepath.
        extension: File extension.

    Returns:
        File with extension.

    Raises:
        TypeError: when `filename` is not type `Path` or `str`.
        TypeError: when `extension` is not type `str`.

    """
    if not isinstance(__file, (PurePath, str)):
        message = f"expected type 'Path' or 'str', got {type(__file)} instead"
        raise TypeError(message)

    if isinstance(__file, PurePath):
        return set_extension_on_filepath(__file, extension)

    if isinstance(__file, str):  # type: ignore
        return set_extension_on_filename(__file, extension)


def set_extension_on_filename(filename: str, extension: str) -> str:
    """Set extension on filename.

    Args:
        filename: Name of file.
        extension: File extension.

    Returns:
        Filename with extension.

    Raises:
        TypeError: when `filename` is not type `str`.
        TypeError: when `extension` is not type `str`.

    """
    if not isinstance(filename, str):  # type: ignore
        message = f"expected type 'str', got {type(filename)} instead"
        raise TypeError(message)

    if not isinstance(extension, str):  # type: ignore
        message = f"expected type 'str', got {type(extension)} instead"
        raise TypeError(message)

    standardized_extension = standardize_file_extension(extension)
    result = (
        f"*{filename!s}*.{standardized_extension!s}"
        if not filename.endswith(f".{standardized_extension!s}")
        else f"*{filename!s}"
    )
    return result


def set_extension_on_filepath(filepath: PurePath, extension: str) -> PurePath:
    """Set extension on filepath.

    Args:
        filepath: Path to file.
        extension: File extension.

    Returns:
        Filepath with extension.

    Raises:
        TypeError: when `filepath` is not type `Path`.
        TypeError: when `extension` is not type `str`.

    """
    if not isinstance(filepath, PurePath):  # type: ignore
        message = f"expected type 'Path', got {type(filepath)} instead"
        raise TypeError(message)

    if not isinstance(extension, str):  # type: ignore
        message = f"expected type 'str', got {type(extension)} instead"
        raise TypeError(message)

    ext = standardize_file_extension(extension)
    filename = ".".join([filepath.name, ext])
    result = filepath.parent / filename
    return result


def standardize_file_extension(extension: str) -> str:
    """Standardize file extension.

    Args:
        extension: File extension.

    Returns:
        File extension.

    Raises:
        TypeError: when `extension` is not type `str`.

    """
    if not isinstance(extension, str):  # type: ignore
        message = f"expected type 'str', got {type(extension)} instead"
        raise TypeError(message)

    result = extension.lower().strip(".")
    return result



# -*- coding: utf-8 -*-

# Standard Library Imports
import abc
import datetime
from typing import Any
from typing import Type
from typing import TypeVar

__all__ = ["MessageMeta"]


# Custom types
T = TypeVar("T")


class MessageMeta(abc.ABCMeta):
    """Implements a message metaclass."""

    def __call__(cls: Type[T], *args: Any, **kwargs: Any) -> T:
        instance = super().__call__(*args, **kwargs)
        setattr(instance, "__created_at__", datetime.datetime.now())
        return instance



# -*- coding: utf-8 -*-
"""Distance Converter.

Module provides functions for converting values to distances.

"""

# Standard Library Imports
import decimal
import re
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_distance"]


def to_distance(__value: Any, /, default: float = 0.0) -> float:
    """Convert value to distance.

    Args:
        __value: Value to convert to distance.
        default (optional): Default value. Default ``0.0``.

    Returns:
        Distance.

    """
    converter = DistanceConverter(default=default)
    result = converter(__value)
    return result


class DistanceConverter(BaseConverter):
    """Class implements a distance converter.

    Args:
        default (optional): Default value. Default ``0.0``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: float = 0.0,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if not isinstance(default, float):
            message = f"expected type 'float', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()


def distance_from_decimal(__value: decimal.Decimal, _: float, /) -> float:
    """Convert decimal value to distance.

    Args:
        __value: Value to convert to distance.

    Returns:
        Distance.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def distance_from_float(__value: float, _: float, /) -> float:
    """Convert float value to distance.

    Args:
        __value: Value to convert to distance.

    Returns:
        Distance.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def distance_from_int(__value: int, _: float, /) -> float:
    """Convert integer value to distance.

    Args:
        __value: Value to convert to distance.

    Returns:
        Distance.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = float(__value)
    return result


def distance_from_str(__value: str, default: float = 0.0, /) -> float:
    """Convert string value to distance.

    Args:
        __value: String representation of distance value.

    Returns:
        Distance.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to distance.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    if not value:
        return default

    try:
        representation = re.sub(r"[^0-9a-zA-Z.]+", r"", value)
        number = re.sub(r"(\d+)\s?m?", r"\1", representation, flags=re.I)
        result = float(number)

    except ValueError:
        message = f"{type(__value)} cannot be converted to distance"
        raise ValueError(message)

    else:
        return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., float]] = {
    decimal.Decimal: distance_from_decimal,
    float: distance_from_float,
    int: distance_from_int,
    str: distance_from_str,
}



# -*- coding: utf-8 -*-
"""Integer Converter.

Module provides function for converting values to integers.

"""

# Standard Library Imports
import datetime
import decimal
import time
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_integer"]


def to_integer(__value: Any, /, default: int = 0) -> int:
    """Convert value to integer.

    Args:
        __value: Value to convert to integer.
        default (optional): Default value. Default ``0``.

    Returns:
        Integer.

    """
    converter = IntegerConverter(default=default)
    result = converter(__value)
    return result


class IntegerConverter(BaseConverter):
    """Class implements a integer converter.

    Args:
        default (optional): Default value. Default ``0``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: int = 0,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if not isinstance(default, int):  # type: ignore
            message = f"expected type 'int', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()


def int_from_bool(__value: bool, _: int, /) -> int:
    """Convert boolean value to ``int``.

    Args:
        __value: Value to convert to ``int``.

    Returns:
        Integer.

    Raises:
        TypeError: when value is not type 'bool'.

    """
    if not isinstance(__value, bool):  # type: ignore
        message = f"expected type 'bool', got {type(__value)} instead"
        raise TypeError(message)

    result = int(__value)
    return result


def int_from_date(__value: datetime.date, _: int, /) -> int:
    """Convert date value to ``int``.

    Args:
        __value: Value to convert to ``int``.

    Returns:
        Integer.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    timestamp = time.mktime(__value.timetuple())
    result = int(timestamp)
    return result


def int_from_datetime(__value: datetime.datetime, _: int, /) -> int:
    """Convert datetime value to ``int``.

    Args:
        __value: Value to convert to ``int``.

    Returns:
        Integer.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    timestamp = __value.timestamp()
    result = int(timestamp)
    return result


def int_from_decimal(__value: decimal.Decimal, _: int, /) -> int:
    """Convert decimal value to ``int``.

    Args:
        __value: Value to convert to ``int``.

    Returns:
        Integer.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = int(__value)
    return result


def int_from_float(__value: float, _: int, /) -> int:
    """Convert float value to ``int``.

    Args:
        __value: Value to convert to ``int``.

    Returns:
        Integer.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = int(__value)
    return result


def int_from_int(__value: int, _: int, /) -> int:
    """Convert integer value to ``int``.

    Args:
        __value: Value to convert to ``int``.

    Returns:
        Integer.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = int(__value)
    return result


def int_from_str(__value: str, default: int = 0, /) -> int:
    """Convert string value to ``int``.

    Args:
        __value: Value to convert to ``int``.
        default (optional): Default value. Default ``0``.

    Returns:
        Integer.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to ``int``.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    if not value:
        return default

    try:
        result = int(float(value))
    except ValueError:
        message = f"{__value} cannot be converted to int"
        raise ValueError(message)

    return result


DEFAULT_CONVERSIONS: Dict[type, Callable[..., int]] = {
    bool: int_from_bool,
    datetime.date: int_from_date,
    datetime.datetime: int_from_datetime,
    decimal.Decimal: int_from_decimal,
    float: int_from_float,
    int: int_from_int,
    str: int_from_str,
}



# -*- coding: utf-8 -*-

# Standard Library Imports
import importlib

# Local Imports
from .abstract_file_wrappers import *
from .csv_file_wrappers import *
from .json_file_wrappers import *
from .txt_file_wrappers import *

try:
    importlib.import_module("openpyxl")
except (ImportError, ModuleNotFoundError):
    pass
else:
    from .xlsx_file_wrappers import *



# -*- coding: utf-8 -*-
"""Boolean Converter.

Module provides function for converting values to booleans.

"""

# Standard Library Imports
import datetime
import decimal
from typing import Any
from typing import Callable
from typing import Dict
from typing import Literal

# Local Imports
from .base_converter import BaseConverter

__all__ = ["to_boolean"]


# Constants
FALSY_VALUES = ("false", "no", "n", "0")
TRUTHY_VALUES = ("true", "yes", "y", "1")


def to_boolean(__value: Any, /, default: bool = False) -> bool:
    """Convert value to boolean.

    Args:
        __value: Value to convert to boolean.
        default (optional): Default value. Default ``False``.

    Returns:
        Boolean.

    """
    converter = BooleanConverter(default=default)
    result = converter(__value)
    return result


class BooleanConverter(BaseConverter):
    """Class implements a boolean converter.

    Args:
        default (optional): Default value. Default ``False``.
        on_error (optional): Whether to raise error or return default. Default ``raise``.

    """

    def __init__(
        self,
        *,
        default: bool = False,
        on_error: Literal["default", "raise"] = "raise",
    ) -> None:
        if not isinstance(default, bool):  # type: ignore
            message = f"expected type 'bool', got {type(default)} instead"
            raise TypeError(message)

        super().__init__(default=default, on_error=on_error)
        self._conversions.update(DEFAULT_CONVERSIONS)
        self._conversions = self._conversions.new_child()


def bool_from_bool(__value: bool, _: bool, /) -> bool:
    """Convert boolean value to ``bool``.

    Args:
        __value: Value to convert to ``bool``.

    Returns:
        Boolean.

    Raises:
        TypeError: when value is not type 'bool'.

    """
    if not isinstance(__value, bool):  # type: ignore
        message = f"expected type 'bool', got {type(__value)} instead"
        raise TypeError(message)

    result = bool(__value)
    return result


def bool_from_date(__value: datetime.date, _: bool, /) -> bool:
    """Convert date value to ``bool``.

    Args:
        __value: Value to convert to ``bool``.

    Returns:
        Boolean.

    Raises:
        TypeError: when value is not type 'date'.

    """
    if not isinstance(__value, datetime.date):  # type: ignore
        message = f"expected type 'date', got {type(__value)} instead"
        raise TypeError(message)

    result = bool(__value)
    return result


def bool_from_datetime(__value: datetime.datetime, _: bool, /) -> bool:
    """Convert datetime value to ``bool``.

    Args:
        __value: Value to convert to ``bool``.

    Returns:
        Boolean.

    Raises:
        TypeError: when value is not type 'datetime'.

    """
    if not isinstance(__value, datetime.datetime):  # type: ignore
        message = f"expected type 'datetime', got {type(__value)} instead"
        raise TypeError(message)

    result = bool(__value)
    return result


def bool_from_decimal(__value: decimal.Decimal, _: bool, /) -> bool:
    """Convert decimal value to ``bool``.

    Args:
        __value: Value to convert to ``bool``.

    Returns:
        Boolean.

    Raises:
        TypeError: when value is not type 'Decimal'.

    """
    if not isinstance(__value, decimal.Decimal):  # type: ignore
        message = f"expected type 'Decimal', got {type(__value)} instead"
        raise TypeError(message)

    result = bool(__value)
    return result


def bool_from_float(__value: float, _: bool, /) -> bool:
    """Convert float value to ``bool``.

    Args:
        __value: Value to convert to ``bool``.

    Returns:
        Boolean.

    Raises:
        TypeError: when value is not type 'float'.

    """
    if not isinstance(__value, float):
        message = f"expected type 'float', got {type(__value)} instead"
        raise TypeError(message)

    result = bool(__value)
    return result


def bool_from_int(__value: int, _: bool, /) -> bool:
    """Convert integer value to ``bool``.

    Args:
        __value: Value to convert to ``bool``.

    Returns:
        Boolean.

    Raises:
        TypeError: when value is not type 'int'.

    """
    if not isinstance(__value, int):  # type: ignore
        message = f"expected type 'int', got {type(__value)} instead"
        raise TypeError(message)

    result = bool(__value)
    return result


def bool_from_str(__value: str, default: bool = False, /) -> bool:
    """Convert string value to `bool`.

    Args:
        __value: Value to convert to ``bool``.
        default (optional): Default value. Default ``False``.

    Returns:
        Boolean.

    Raises:
        TypeError: when value is not type 'str'.
        ValueError: when value cannot be converted to ``bool``.

    """
    if not isinstance(__value, str):  # type: ignore
        message = f"expected type 'str', got {type(__value)} instead"
        raise TypeError(message)

    value = __value.replace("  ", " ").strip()
    if not value:
        return default

    if value.lower() in TRUTHY_VALUES:
        return True

    if value.lower() in FALSY_VALUES:
        return False

    raise ValueError(f"'{__value}' cannot be converted to bool")


DEFAULT_CONVERSIONS: Dict[type, Callable[..., bool]] = {
    bool: bool_from_bool,
    datetime.date: bool_from_date,
    datetime.datetime: bool_from_datetime,
    decimal.Decimal: bool_from_decimal,
    float: bool_from_float,
    int: bool_from_int,
    str: bool_from_str,
}



# -*- coding: utf-8 -*-
"""Settings."""

# File Extensions
CSV_EXTENSION = ".csv"
JSON_EXTENSION = ".json"
PDF_EXTENSION = ".pdf"
TXT_EXTENSION = ".txt"
XLSX_EXTENSION = ".xlsx"
ZIP_EXTENSION = ".zip"

# Defaults
DEFAULT_CSV_DELIMITER = ","
DEFAULT_CSV_DIALECT = "excel"
DEFAULT_CSV_NEWLINE = ""
DEFAULT_CSV_QUOTECHAR = '"'
DEFAULT_FILE_ENCODING = "utf-8"
DEFAULT_TXT_NEWLINE = ""



# -*- coding: utf-8 -*-
"""Base Dispatcherr."""

# Local Imports
from .abstract_dispatcher import AbstractDispatcher
from ..messagebus import AbstractMessageBus

__all__ = ["BaseDispatcher"]


class BaseDispatcher(AbstractDispatcher):
    """Implements a base class for dispatchers to inherit.

    Args:
        __bus: Message bus.

    Attributes:
        messagebus: Message bus.

    Raises:
        TypeError: when argument is not type ``AbstractMessageBus``.

    """

    def __init__(self, __bus: AbstractMessageBus, /) -> None:
        if not isinstance(__bus, AbstractMessageBus):  # type: ignore
            expected = "expected type 'AbstractMessageBus'"
            actual = f"got {type(__bus)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        self._messagebus = __bus

    @property
    def messagebus(self) -> AbstractMessageBus:
        """Message bus."""
        return self._messagebus



# -*- coding: utf-8 -*-
"""File-System Repository."""

# Standard Library Imports
from typing import Any

# Local Imports
from .abstract_repository import AbstractRepository
from ..wrappers import AbstractFileSystemWrapper

__all__ = ["AbstractFileSystemRepository"]


class AbstractFileSystemRepository(AbstractRepository):
    """Represents an abstract file-system repository.

    Args:
        __file: File.

    """

    def __init__(self, __file: Any, /, *args: Any, **kwargs: Any) -> None:
        if not isinstance(__file, AbstractFileSystemWrapper):
            expected = "expected type 'AbstractFileSystemWrapper'"
            actual = f"got {type(__file)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        super().__init__(*args, **kwargs)
        self._file = __file



# -*- coding: utf-8 -*-
"""Message Bus.

This module defines message-bus classes which delivers commands and events
to their respective handlers.

Commands are simple data structures that capture an intent for the system to
perform a particular action: commands are always matched to a single handler.
When a command is executed, there is an expectation that an event will occur
as a result; whenever a process fails, the process or user that created the
command must receive an error message containing pertinent information.

Events are data structures that are broadcast to all subscribed listeners.
Events are never assigned to dedicated event handlers. Instead, handlers are
registered with the message bus and wait for an event to occur before
performing a particular action. Events reflect business logic described in
terms of 'if this happens, then do that'. They are used to implement workflows
in the system.

Implementation based on 'Architecture Patterns in Python' message-bus pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# Standard Library Imports
from __future__ import annotations
import abc
import logging
from operator import methodcaller
from typing import Any
from typing import Callable
from typing import Dict
from typing import List
from typing import Literal
from typing import Optional
from typing import Type

# Local Imports
from .messages import BaseCommand
from .messages import BaseEvent
from .messages import AbstractMessage
from .queues import MessageQueue
from .units_of_work import AbstractUnitOfWork
from . import config

__all__ = [
    "AbstractMessageBus",
    "MessageBus",
]


# Initialize logger.
log = logging.getLogger("dodecahedron")

# Custom types
Handler = Callable[..., None]
CommandHandlers = Dict[Type[BaseCommand], Handler]
EventHandlers = Dict[Type[BaseEvent], List[Handler]]

# Constants
eventcollector = methodcaller("collect_events")
IGNORE = "ignore"
RAISE = "raise"


class AbstractMessageBus(abc.ABC):
    """Represents an abstract message bus.

    Attributes:
        uow: Unit of work.
        queue: Message queue.

    """

    @property
    @abc.abstractmethod
    def uow(self) -> AbstractUnitOfWork:
        """Unit of work."""
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def queue(self) -> MessageQueue:
        """Message queue."""
        raise NotImplementedError

    @abc.abstractmethod
    def handle(
        self,
        message: AbstractMessage,
        /,
        callback: Optional[Callable[..., Any]] = None,
    ) -> None:
        """Handle a message.

        Args:
            message: Message to handle.
            callback (optional): Function to call after handling message. Default ``None``.

        """
        raise NotImplementedError

    @abc.abstractmethod
    def subscribe(
        self, message: Type[AbstractMessage], handler: Handler
    ) -> None:
        """Subscribe a handler for a `Command` or `Event`.

        Args:
            message: Message type to which to subscribe.

        """
        raise NotImplementedError


class MessageBus(AbstractMessageBus):
    """Implements a message buse.

    Args:
        uow: Unit of work.
        command_handlers: Dictionary of commands and handlers.
            Each command is associated with a single handler.
        event_handlers: Dictionary of events and handlers.
            Each event is associated with multiple handlers.

    Attributes:
        uow: Unit of work.
        queue: Message queue.

    """

    def __init__(
        self,
        uow: AbstractUnitOfWork,
        /,
        command_handlers: CommandHandlers,
        event_handlers: EventHandlers,
    ) -> None:
        self._uow = uow
        self._command_handlers = command_handlers
        self._event_handlers = event_handlers
        self._queue = MessageQueue()

    @property
    def uow(self) -> AbstractUnitOfWork:
        """Unit of work."""
        return self._uow

    @property
    def queue(self) -> MessageQueue:
        """Message queue."""
        return self._queue

    def handle(
        self,
        message: AbstractMessage,
        /,
        callback: Optional[Callable[..., Any]] = None,
    ) -> None:
        """Handle message.

        Provided message is passed to an appropriate handler function.

        Args:
            message: Message.
            callback (optional): Function to call after handling message. Default ``None``.

        """
        if not isinstance(message, AbstractMessage):  # type: ignore
            expected = "expected type 'AbstractMessage'"
            actual = f"got {type(message)} instead"
            error = ", ".join([expected, actual])
            raise TypeError(error)

        self.queue.append(message)
        while self.queue:
            message = self.queue.popleft()
            self._handle_message(message)

        if callback is not None:
            callback()

    def subscribe(
        self,
        message: Type[AbstractMessage],
        handler: Handler,
    ) -> None:
        """Subscribe handler to `Command` or `Event`.

        Args:
            message: Message type to which to subscribe.
            handler: Handler function to subscribe.

        """
        if not issubclass(message, (BaseCommand, BaseEvent)):
            msg = f"{type(message)} was not a 'Command' or an 'Event'"
            raise TypeError(msg)

        if issubclass(message, BaseCommand):
            self._command_handlers[message] = handler

        if issubclass(message, BaseEvent):
            self._event_handlers.setdefault(message, [])
            self._event_handlers[message].append(handler)

    def _handle_message(self, message: AbstractMessage, /) -> None:
        """Handle message.

        Args:
            message: Message to handle.

        """
        if not isinstance(message, AbstractMessage):  # type: ignore
            expected = "expected type 'AbstractMessage'"
            actual = f"got {type(message)} instead"
            error = ", ".join([expected, actual])
            raise TypeError(error)

        if not isinstance(message, (BaseCommand, BaseEvent)):
            msg = f"{type(message)} was not a 'Command' or an 'Event'"
            raise TypeError(msg)

        if isinstance(message, BaseCommand):
            self._pass_message_to_command_handler(message, on_error=RAISE)

        if isinstance(message, BaseEvent):
            self._pass_message_to_event_handlers(message, on_error=IGNORE)

    def _pass_message_to_command_handler(
        self,
        message: AbstractMessage,
        /,
        on_error: Literal["ignore", "raise"] = RAISE,
    ) -> None:
        """Pass message to command handler.

        Args:
            message: Message to handle.
            on_error (optional): Strategy for handling errors. Default ``raise``.

        """
        if not isinstance(message, BaseCommand):  # type: ignore
            expected = "expected type 'BaseCommand'"
            actual = f"got {type(message)} instead"
            msg = ", ".join([expected, actual])
            raise TypeError(msg)

        try:
            handler = self._command_handlers[type(message)]
            log.debug("Handling '%s' with handler %s", message, handler)
            handler(message)

        except Exception as error:
            handle_error(message, error, on_error=on_error)

        self._collect_events()

    def _pass_message_to_event_handlers(
        self,
        message: AbstractMessage,
        /,
        on_error: Optional[Literal["ignore", "raise"]] = IGNORE,
    ) -> None:
        """Pass message to event handlers.

        Args:
            message: Message to handle.
            on_error (optional): Strategy for handling errors. Default ``ignore``.

        """
        if not isinstance(message, BaseEvent):  # type: ignore
            expected = "expected type 'BaseEvent'"
            actual = f"got {type(message)} instead"
            msg = ", ".join([expected, actual])
            raise TypeError(msg)

        for handler in self._event_handlers[type(message)]:
            try:
                log.debug("Handling '%s' with handler %s", message, handler)
                handler(message)

            except Exception as error:
                handle_error(message, error, on_error=on_error)

            self._collect_events()

    def _collect_events(self) -> None:
        """Collect events."""
        if hasattr(self.uow, "collect_events"):
            events = list(eventcollector(self.uow))
            self.queue.extend(events)


# ----------------------------------------------------------------------------
# Helpers
# ----------------------------------------------------------------------------
def handle_error(
    message: AbstractMessage,
    error: Exception,
    *,
    on_error: Optional[Literal["ignore", "raise"]],
) -> None:
    """Handle error raised when handling message.

    Args:
        message: Message.
        error: Error raised when handling message.
        on_error: Strategy for handling errors.

    """
    exc_info = not config.is_production_environment()
    log.error("Error handling %s", message, exc_info=exc_info)
    if on_error == RAISE:
        raise error



# -*- coding: utf-8 -*-
"""Abstract Unit of Work.

Based on 'Architecture Patterns in Python' unit-of-work pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# Standard Library Imports
from __future__ import annotations
import abc
from typing import Any
from typing import Optional
from typing import Type

__all__ = ["AbstractUnitOfWork"]


# Constants
AUTO_COMMIT_ATTR = "_auto_commit"


class AbstractUnitOfWork(abc.ABC):
    """Represents an abstract unit of work."""

    @property
    def auto_commit(self) -> bool:
        """Whether to auto commit.

        Raises:
            TypeError: when value is not type 'bool'.

        """
        result = getattr(self, AUTO_COMMIT_ATTR, False)
        return result

    @auto_commit.setter
    def auto_commit(self, value: bool) -> None:
        if not isinstance(value, bool):  # type: ignore
            message = f"expected type 'bool', got {type(value)} instead"
            raise TypeError(message)

        setattr(self, AUTO_COMMIT_ATTR, value)

    def __enter__(self) -> AbstractUnitOfWork:
        return self

    def __exit__(self, exc: Optional[Type[Exception]], *_: Any) -> None:
        if self.auto_commit is True and not exc:
            self.commit()

    def commit(self) -> None:
        """Commit changes."""

    def rollback(self) -> None:
        """Rollback changes."""



# -*- coding: utf-8 -*-
"""Abstract Dispatcherr."""

# Standard Library Imports
import abc

# Local Imports
from ..messagebus import AbstractMessageBus

__all__ = ["AbstractDispatcher"]


class AbstractDispatcher(abc.ABC):
    """Represents an abstract dispatcherr.

    Attributes:
        messagebus: Message bus.

    """

    @property
    @abc.abstractmethod
    def messagebus(self) -> AbstractMessageBus:
        """Message bus."""
        raise NotImplementedError



# -*- coding: utf-8 -*-
"""Abstract models.

This module defines abstract base classes for domain models.

Implementation based on 'Architecture Patterns in Python' domain model pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# pylint: disable=too-few-public-methods

# Standard Library Imports
import abc
from typing import Deque
from typing import Union
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ..messages import AbstractMessage
    from ..queues import MessageQueue

__all__ = [
    "AbstractModel",
    "AbstractAggregate",
]


class AbstractModel(abc.ABC):
    """Represents an abstract model.

    Models have one responsibility: to be unique. Therefore, subclasses must
    implement both the `__eq__` and `__hash__` methods.

    """

    @abc.abstractmethod
    def __eq__(self, other: object) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def __hash__(self) -> int:
        raise NotImplementedError


class AbstractAggregate(AbstractModel):
    """Represents an abstract aggregate.

    The primary purpose of an aggregate is not simply to hold a collection of
    objects; instead, the purpose of an aggregate is to record events raised
    by the domain model. In addition, the aggregate encapsulates whatever
    business logic is involved when adding and removing objects.

    Attributes:
        events: Events raised by the domain model.

    """

    @property
    @abc.abstractmethod
    def events(self) -> Union[Deque["AbstractMessage"], "MessageQueue"]:
        """Events raised."""
        raise NotImplementedError

    @abc.abstractmethod
    def __contains__(self, obj: object) -> bool:
        raise NotImplementedError

    @abc.abstractmethod
    def add(self, obj: object) -> None:
        """Add object to aggregate."""
        raise NotImplementedError

    @abc.abstractmethod
    def get(self, ref: str) -> object:
        """Get object in aggregate."""
        raise NotImplementedError

    @abc.abstractmethod
    def remove(self, obj: object) -> None:
        """Remove object from aggregate."""
        raise NotImplementedError



# -*- coding: utf-8 -*-

# Local Imports
from .boolean_converter import *
from .currency_converter import *
from .date_converter import *
from .datetime_converter import *
from .distance_converter import *
from .file_mode_converters import *
from .integer_converter import *
from .string_converter import *
from .timestamp_converter import *
from .yes_or_no_converters import *



# -*- coding: utf-8 -*-
"""CSV Repository.

Implementation based on 'Architecture Patterns in Python' repository pattern.

.. _Architecture Patterns in Python:
    https://github.com/cosmicpython/code

"""

# Standard Library Imports
import logging
from typing import Any
from typing import Optional
from typing import Sequence

# Local Imports
from .file_system_repository import AbstractFileSystemRepository
from ..wrappers.csv_file_wrappers import AbstractCsvWrapper

__all__ = ["AbstractCsvRepository"]


# Initiate logger.
log = logging.getLogger("dodecahedron")


class AbstractCsvRepository(AbstractFileSystemRepository):
    """Represents an abstract `.csv` file repository.

    Args:
        __file: CSV file.

    Attributes:
        columns: Columns names.

    """

    def __init__(self, __file: Any, /, *args: Any, **kwargs: Any) -> None:
        if not isinstance(__file, AbstractCsvWrapper):
            expected = "expected type 'AbstractCsvWrapper'"
            actual = f"got {type(__file)} instead"
            message = ", ".join([expected, actual])
            raise TypeError(message)

        super().__init__(__file, *args, **kwargs)

    @property
    def columns(self) -> Optional[Sequence[Any]]:
        """Column names."""
        results = getattr(self._file, "fieldnames", None)
        return results



# -*- coding: utf-8 -*-
"""File Mode Converters."""

# Standard Library Imports
import re

__all__ = [
    "to_bytes_file_mode",
    "to_text_file_mode",
]


def to_bytes_file_mode(mode: str) -> str:
    """Convert to bytes file mode.

    Args:
        mode: Mode.

    Returns:
        Mode.

    """
    result = re.sub(r"[bt]?([arw])[bt]?", r"\g<1>b", mode)
    return result


def to_text_file_mode(mode: str) -> str:
    """Convert to text file mode.

    Args:
        mode: Mode.

    Returns:
        Mode.

    """
    result = re.sub(r"[bt]?([arw])[bt]?", r"\g<1>t", mode)
    return result
